# initialize.py
# collected and adapted by Andre Schneider (S1) 2018/05


# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import matplotlib.pyplot as plt
import logging
import numpy as np
from qkit.measure.timedomain import gate_func
from qkit.measure.timedomain import sequence_library as sl
from qkit.measure.timedomain import VirtualAWG as virtual_awg
import ipywidgets as widgets

class InitializeTimeDomain(object):
    def __init__(self,sample):
        self._sample = sample
        
        self._vawg = virtual_awg.VirtualAWG(self._sample)
        
    params = { #typical values are given in (brackets)
            'T_rep':                " repetition rate [s] (200e-6)",
            'exc_T':                " maximum window for manipulation [s]  (20e-6)",
            'readout_tone_length':  " length of readout tone [s]  (400e-09)",

            'clock':                " samplerate for manipulation awg [S/s] (1.2e9)",
            'readout_clock':        " samplerate for readout dac [S/s] (1.2e9)",
            'spec_samplerate':      " samplerate for readout adc [S/s] (1.25e9)",

            'f01':                  " qubit transition frequency [Hz] (5.469e9)",
            'fr':                   " resonator readout frequency [Hz] (9e9)",
            'iq_frequency':         " IQ frequency for manipulation [Hz] (80e6)",

            'mw_power':             " power for the qubit microwave source",
            
            'awg':                  " Instrument: qubit manipulation awg",
            'qubit_mw_src':         " Instrument: qubit manipulation microwave source",
            'readout_awg':          " Instrument: readout awg",
            'readout_mw_src':       " Instrument: readout microwave source",
            'mspec':                " Instrument virtual_measure_spec"
    }
    optional_params = {
            'readout_iq_frequency': [" IQ frequency for readout [Hz]", 30e6],
            'acqu_window':          [" position of the adc recording window as array [start, end]. No longer needed to be divisible by 32. (samples)", [96, 545]],
            'overlap':              [" Overlap between manipulation window and readout pulse.", 0],
            'readout_delay':        [" Time offset of the readout pulse to adjust the timing of the devices", 0],
            'readout_pha':          [" Phase of the readout tone", 0],
            'readout_relamp':       [" relative amplitude of the different readout tones" , 1],
            'spec_input_level':     [" Input range in mV of the spectrum adc card. Restricted to card-dependant values", 250]
    }

    def initialize(self):
        '''
        doc-string tbd

        you will need a valid sample file with the following parameters set:
         
        '''
        
        # Check if we have all necessary parameters in our sample object:
        breakpoint = False
        for p in self.params:
            if p not in self._sample.__dict__:
                logging.error("Please specify '"+p+"' in your sample object:"+self.params[p])
                breakpoint = True
        if breakpoint: raise ValueError("Not all vallues in your sample file are present. I can not continue. Sorry.")

        # If some parameters are not given, that have a good default, set them in the sample object, so that the user knows them.
        for p,v in self.optional_params.iteritems():
            if p not in self._sample.__dict__:
                self._sample.__dict__[p] = v[1]
        #Init the spectrum card
        self._sample.mspec.spec_stop()
        self._sample.mspec.set_segments(1)
        self._sample.mspec.set_blocks(1)
        self._sample.mspec.set_window(0,512)

        if self._sample.readout_mw_src.has_parameter('high_power'):
            self._sample.readout_mw_src.set_high_power(True)
        
        self.update_qubit_mw_src()
        
        self._sample.readout_mw_src.set_power(15)
        self._sample.qubit_mw_src.set_power(0)
        self._sample.readout_mw_src.set_status(True)
        self._sample.qubit_mw_src.set_status(True)

        #qubit awg presets
        if True: #Tabor AWG in 4ch mode
            # Tabor is self triggered
            # SYNC output of Tabor makes sync pulse to Tek
            # TekMarker makes sync pulse for spec card #CHECK
            self._sample.awg.reset()
            #self.set_p1_runmode('SEQ')
            #self.set_p2_runmode('USER')
            for i in range(1,5):
                self._sample.awg.set('ch%i_amplitude'%i,2)
                self._sample.awg.set('ch%i_output'%i,True)
                self._sample.awg.set('ch%i_marker_output'%i,True)
                self._sample.awg.set('ch%i_marker_high'%i,1)
            self._sample.awg.get_all()
            self._sample.awg.set_common_clock(True)
            self._sample.awg.set_clock(self._sample.clock)
            self._sample.readout_awg.set_clock(self._sample.readout_clock)
            self._load_awg_square([100e-9])
            
            if self._sample.awg._numchannels == 4:
                self._sample.awg.set_p1_trigger_time(self._sample.T_rep)
                self._sample.awg.set_p1_sync_output(True)
            else:
                self._sample.awg.set_trigger_time(self._sample.T_rep)
                self._sample.awg.set_sync_output(True)
            if self._sample.T_rep < 1.5 * self._sample.exc_T:
                raise ValueError("Your repetition rate T_rep is too small for the chosen window exc_T.")
            
            if  "Tektronix" in self._sample.readout_awg.get_type() and self._sample.T_rep < 170e-6:
                raise ValueError("You are using the Tektronix AWG as readout. This has a bug and your repetition rate should be >= 200us.")

            for i in range(1,3):
                self._sample.awg.set('p%i_runmode'%i,'USER')   
                self._sample.awg.set('p%i_trigger_mode'%i,'TRIG')
                self._sample.awg.set('p%i_trigger_source'%i,'TIM')
                self._sample.awg.set('p%i_sequence_mode'%i,'STEP')
                self._sample.awg.set('p%i_trigger_delay'%i,0)
                self._sample.awg.set('p%i_runmode'%i,'SEQ')   
            
        #put something into the manipulation awg
        
        self._sample.readout.set_tone_amp(1)
        self._sample.readout.set_LO(np.mean(self._sample.fr)-self._sample.readout_iq_frequency)   #set LO lower and use mixer as a up-converter. For multiplexing set LO to centerfreq - IQ freq
        self._sample.readout.set_tone_freq(np.atleast_1d(self._sample.fr))   #probe tone frequency
        self._sample.readout.set_tone_pha(np.atleast_1d(self._sample.readout_pha))   #phase shift (usually not necessary)
        self._sample.readout.set_tone_relamp(np.atleast_1d(self._sample.readout_relamp))   #relative amplitude of tone

        self.update_timings()

        self._sample.mspec.set_gate_func(gate_func.Gate_Func(awg = self._sample.awg,sample=self._sample).gate_fcn)
        self._sample.mspec.spec_stop()   #stop card before modifying a setting
        self._sample.mspec.set_samplerate(self._sample.spec_samplerate)   
        self._sample.spec_samplerate = self._sample.mspec.get_samplerate() #We want to have the actual samplerate in the sample object

        self.set_spec_input_level(self._sample.spec_input_level)


    def update_timings(self, zero_dac_delay = False):
        self._sample.awg.set('p1_trigger_time',self._sample.T_rep)
        self._sample.awg.set('trigger_time',self._sample.T_rep)
        self._sample.mspec.set_trigger_rate(1/float(self._sample.T_rep))
        self._sample.readout.set_dac_duration(self._sample.readout_tone_length) #length of the readout tone
        self._sample.readout.set_dac_clock(self._sample.clock)
        self._sample.readout.set_dac_delay(0)
        self._sample.readout.update()

    def set_spec_input_level(self,level):
        '''
        possible levels depend on the card and are specified in mV.
        '''
        self._sample.mspec.spec_stop() 
        self._sample.mspec.set_input_amp(level)

    def record_single_trace(self):
        #record single probe tone signal and plot
        self._sample.qubit_mw_src.set_status(0)
        self._sample.mspec.set_window(0,1024)
        self._sample.mspec.set_averages(1)
        self._sample.mspec.set_segments(5)
        samples = self._sample.mspec.acquire()
        plt.figure(figsize=(15,5))
        plt.plot(samples[:,:,0])
        plt.xticks(np.arange(0,self._sample.mspec.get_samples(),32))
        plt.xlim((0,self._sample.mspec.get_samples()))
        plt.grid()
        plt.xlabel('samples (%.0fMHz samplerate: 1sample = %.3gns)'%(self._sample.mspec.get_samplerate()/1e6,1./self._sample.mspec.get_samplerate()*1e9))
        plt.ylabel('amplitude')
        plt.twiny()
        plt.xlabel('nanoseconds')
        plt.xticks(np.arange(0,float(self._sample.mspec.get_samples())/self._sample.mspec.get_samplerate()*1e9,100))
        plt.xlim(0,float(self._sample.mspec.get_samples())/self._sample.mspec.get_samplerate()*1e9)
        plt.ylim(-128,127)
        clips=np.size(np.where(np.array(samples).flatten()==127))+np.size(np.where(np.array(samples).flatten()==-128))
        if clips > 50:
            logging.error("Clipping detected, please reduce amplifier voltage (%r Clips)" %clips)
        elif clips > 0:
            print("In 5 measurements, only %i points had the maximum amplitude. This sounds good, go on."%clips)
        elif np.max(samples)<15:
            raise ValueError("No signal detected. Check your wiring!")
        elif np.max(np.abs(samples)) < 64:
            print("The amplitude never reached half of the maximum of the card. Think about adding an amplifier and check that you are using the smallest spec_input_level.")
        else:
            print("The amplitude seems reasonable. Check the image nevertheless.")
        self._sample.mspec.set_window(*self._sample.acqu_window)
        
    def record_averaged_trace(self):
        #record average signal and plot
        self._sample.mspec.spec_stop()
        self._sample.mspec.set_spec_trigger_delay(0)
        self._sample.mspec.set_samples(1024)
        self._sample.mspec.set_averages(1e4)
        self._sample.mspec.set_segments(1)
        plt.figure(figsize=(15,5))
        plt.plot(self._sample.mspec.acquire())
        plt.xlim((0,self._sample.mspec.get_samples()))
        plt.grid()
        plt.xlabel('samples (%.0fMHz samplerate: 1sample = %.3gns)'%(self._sample.mspec.get_samplerate()/1e6,1./self._sample.mspec.get_samplerate()*1e9))
        if self._sample.mspec.get_samples() > 512:
            plt.xticks(np.arange(0,self._sample.mspec.get_samples(),32))
        else:
            plt.xticks(np.arange(0,self._sample.mspec.get_samples(),16))
        plt.twiny()
        plt.xlabel('nanoseconds')
        plt.xticks(np.arange(0,float(self._sample.mspec.get_samples())/self._sample.mspec.get_samplerate()*1e9,100))
        plt.xlim(0,float(self._sample.mspec.get_samples())/self._sample.mspec.get_samplerate()*1e9)
        plt.ylabel('amplitude')
        self._sample.mspec.set_window(*self._sample.acqu_window)
        
    def align_pulses(self, samples=768):
        #record single probe tone signal and plot
        samples = int(samples/32)*16
        self._sample.mspec.spec_stop()
        self._sample.mspec.set_segments(2)
        self._sample.mspec.set_averages(1e4)
        self._sample.mspec.set_spec_trigger_delay(0)
        self._sample.mspec.set_samples(samples*2)
        self._sample.mspec._dacq.set_post_trigger(samples)
            
        self._sample.qubit_mw_src.set_frequency(self._sample.readout_mw_src.get_frequency())
        sr = self._sample.mspec.get_samplerate()/1e9
        pwr = self._sample.qubit_mw_src.get_power()
        self._sample.qubit_mw_src.set_power(5)
        self._load_awg_square([0,50e-9])
        self._sample.qubit_mw_src.set_status(1)
        msp = self._sample.mspec.acquire()
        self._sample.qubit_mw_src.set_status(0)
        plt.figure(figsize=(15,5))
        plt.plot(np.arange(-samples+32,samples+32),msp[:,:,1])
        plt.plot(np.arange(-samples+32,samples+32),msp[:,:,0],'--k')
        plt.xlim((-samples+32,samples+32))
        plt.grid()
        plt.xlabel('samples (%.0fMHz samplerate: 1 sample = %.3gns)'%(sr*1e3,1./sr))
        plt.ylabel('amplitude')
        plt.twiny()
        plt.xlabel('nano seconds')
        duration = float(samples)/sr
        plt.xticks(np.arange(-round(duration,-2),round(duration,-2),100))
        plt.xlim(-duration+32/sr,+duration+32/sr)
        plt.grid(axis='x')
        self._sample.qubit_mw_src.set_power(pwr)
        
        print("You can now change the self._sample.readout_delay parameter to align the two pulses. The readout pulse should start when the manipulation ends.")
        print("If you are satisfied just press <Enter> or tell me the new delay.")
        plt.show()
        inp = raw_input("readout_delay ({:g}s) = ".format(self._sample.readout_delay))
        if inp == '':
            return
        if np.abs(float(inp)) > 1e-3:
            raise ValueError("Your delay is more than a mili second. I guess you are wrong.")
        self._sample.readout_delay = float(inp)
        self.update_timings()
        self.align_pulses()
        
    def crop_recording_window(self):
        self._sample.mspec.spec_stop()
        self._sample.mspec.set_averages(1e4)
        self._sample.mspec.set_window(0,512)
        self._sample.mspec.set_segments(1)
        msp = self._sample.mspec.acquire()
        
        def pltfunc(start,end,done):
            if done:
                self._sample.acqu_window = [start,end]
                self._sample.mspec.set_window(start,end)
                self._sw.disabled = True
                self._ew.disabled = True
                self._dw.disabled = True
                self._dw.description = "acqu_window set to [{:d}:{:d}]".format(start,end)
            else:
                plt.figure(figsize=(15,5))
                plt.plot(msp)
                plt.axvspan(0,start,color='k',alpha=.2)
                plt.axvspan(end,len(msp),color='k',alpha=.2)
                plt.xlim(0,len(msp))
                plt.show()
        self._sw =  widgets.IntSlider(min=0,max=len(msp),step=1,value=self._sample.acqu_window[0],continuous_update=True)
        self._ew = widgets.IntSlider(min=0,max=len(msp),step=1,value=self._sample.acqu_window[1],continuous_update=True)
        self._dw = widgets.Checkbox(value=False,description="Done!",indent=True)
        self._wgt = widgets.interact(pltfunc,start=self._sw,end=self._ew,done=self._dw)
        self._sample.mspec.set_window(*self._sample.acqu_window)
        
    def check_sidebands(self):
        rec = self._sample.readout.spectrum()
        ro = self._sample.readout.readout()
        plt.figure(figsize=(15,5))
        plt.plot(rec[0],rec[1]/len(rec[1]),'--o')
        ylim = plt.ylim()
        plt.vlines(self._sample.readout.get_LO(),*ylim,color='r')
        for i,fr in enumerate(np.atleast_1d(self._sample.fr)):
            plt.plot([fr],ro[0][i],'+',ms=50,mew=3)
        plt.ylim(0,ylim[1])
        spread = np.ptp(np.append(np.atleast_1d(self._sample.fr),self._sample.readout.get_LO()))*1.2
        plt.xlim([self._sample.readout.get_LO()-spread,self._sample.readout.get_LO()+spread])
        plt.grid()

    def update_qubit_mw_src(self):
        self._sample.qubit_mw_src.set_power(self._sample.mw_power)
        self._sample.qubit_mw_src.set_frequency(self._sample.f01 - self._sample.iq_frequency)
        
    def _load_awg_square(self, times):
        self._vawg.set_sequence(sl.rabi(self._sample, iq_frequency=0), times)
        self._vawg.load(show_progress_bar=False)



# -*- coding: utf-8 -*-
"""
@author: MP, TW, HR@KIT/2018
"""
import qkit
import logging
from qkit.core.lib.file_service.file_info_database import fid


def _load_file_service():
    logging.info("loading service: file info database (fid)")
    qkit.fid = fid()
    #info: qkit.store_db does not exist anymore: use qkit.fid instead.

if qkit.cfg.get('fid_scan_datadir', True):
    _load_file_service()



# qtLAB preliminary driver for RF frontend PCB
# filename: RFPCB.py
# Robert Gartmann <gartmann@kit.edu>, 2021
# Marius Frohn <uzrfo@student.kit.edu>, 2022

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
#import zerorpc
#import types
import logging
#import json
import warnings
from time import sleep
import math
#from enum import Enum

import grpc
from proto_gen import hflo_pb2_grpc as HFL_grpc
from proto_gen import hflo_pb2 as HFL
#from proto_gen import lmx2594evm_pb2_grpc as HFL_grpc
#from proto_gen import lmx2594evm_pb2 as HFL
from proto_gen import hfdemodulator_pb2_grpc as HFD_grpc
from proto_gen import hfdemodulator_pb2 as HFD
from proto_gen import hfattenuator_pb2_grpc as HFA_grpc
from proto_gen import hfattenuator_pb2 as HFA
from proto_gen import hfsuperv2_pb2_grpc as SuperV2_grpc
from proto_gen import hfsuperv2_pb2 as SuperV2

class RFPCB_V2(Instrument):
    '''
    This is the pre-release python driver for the RF frontend PCB
    Usage:
        Initialise with
        <name> = instruments.create('<name>', "RFPCB", address='<TCP/IP>:<port>')
        Set/Get Power and Frequency implemented as Qkit Instrument paramters
        Call gRPC functions with 
        <name>.xyz_stub.Function(XYZ.ProtoType(param_name=*value*)) # Dont forget UpdateWrite!
    '''
    def __init__(self, name, address, model = 'RFPCB'):
        '''
        Initializes the ZeroRPC Client to communicate with according Server (RPi).
        Input:
            name (string)    : name of the instrument
            address (string) : Servicehub Address including port e.g. "192.168.99.132:50058" for Platform or "192.168.99.206:50058" for NanoPi
        '''

        # Init Qkit Instrument
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name)
        self._model = model

        # connect to IPESH
        self._channel = grpc.insecure_channel(address)
        self.hfl_stub = HFL_grpc.HfLOServiceStub(self._channel)
        #self.hfl_stub = HFL_grpc.LMX2594EVMServiceStub(self._channel)
        self.hfd_stub = HFD_grpc.HfDemodulatorServiceStub(self._channel)
        self.hfa_stub = HFA_grpc.HfAttenuatorServiceStub(self._channel)
        self.v2_stub = SuperV2_grpc.HfSuperV2ServiceStub(self._channel)

        # default values for Instrument
        self._connection = True
        self._status = False
        self._power = 30

        # Endpoint Index List
        # will be used with '...index=self.EPDict.index("name")...'
        # must be adjusted according to order of eps on devices (check IPESH info)
        self.EPDict = ["TXPL", "RXPL", "IFLO", "RFLO"]

        # Implement parameters
        # Main
        self.add_parameter('frequency', type=float,
            tags=["Main", "Multi"],
            minval=4e9, maxval=12e9, units='Hz')
        self.add_parameter('power', type=float,
            tags=["Main", "Attenuator"],
            minval=0, maxval=30, units='dBm')
        self.add_parameter('status', type=bool, tags=["Main", "Multi"])
        self.add_parameter('connection', type=bool, tags=["Main"])
        # Detail
        self.add_parameter('iflo_frequency', type=float,
            tags=["Detail", "LO"], listen_to=[], # Add frequency once implemented
            minval=9765625.0, maxval=15e9, units='Hz')
        self.add_parameter('rflo_frequency', type=float,
            tags=["Detail", "LO"], listen_to=[], # Add frequency once implemented
            minval=9765625.0, maxval=15e9, units='Hz')
        self.add_parameter('txpl_frequency', type=float,
            tags=["Detail", "LO"], 
            minval=9765625.0, maxval=15e9, units='Hz')
        self.add_parameter('rxpl_frequency', type=float,
            tags=["Detail", "LO"], 
            minval=9765625.0, maxval=15e9, units='Hz')
        self.add_parameter('demod_lo_match', type=int,
            tags=["Detail", "Demodulator"], listen_to=[], # Add frequency,iflo once implemented
            minval=3e8, maxval=6e9, units='Hz')
        self.add_parameter('demod_gain', type=int, 
            tags=["Detail", "Demodulator"], 
            minval=int(0), maxval=int(7), unit='dB')
        self.add_parameter('demod_gain_error', type=float, 
            tags=["Detail", "Demodulator"], 
            minval=-50.0, maxval=50.0, unit='cdB')
        self.add_parameter('demod_attenuation', type=int, 
            tags=["Detail", "Demodulator"], 
            minval=int(0), maxval=int(31), unit='dB')
        self.add_parameter('demod_phase_error', type=float, 
            tags=["Detail", "Demodulator"], 
            minval=-4.14, maxval=1.94, unit='1°')
        self.add_parameter('demod_dc_offset_i', type=float, 
            tags=["Detail", "Demodulator"], 
            minval=-70, maxval=100, unit='mV')
        self.add_parameter('demod_dc_offset_q', type=float, 
            tags=["Detail", "Demodulator"], 
            minval=-70, maxval=100, unit='mV')

    # --- Provide set/get for instrument parameters ---
    # Main
    def do_get_frequency(self, query=True):
        '''
        Get frequency of device
        (So far C++ implementation only returns viable results for mixsweep)
        Input:
            None
        Output:
            freq (float) : Frequency in Hz
        '''
        #return self.v2_stub.GetDeviceFreq(SuperV2.Empty()).value
        return 4e9
    def do_set_frequency(self, frequency, offset = 100e6, method="mixsweep"):
        '''
        Set frequency of device.
        Default BB frequency is 100MHz.
        Default method for sweeping is mixsweep.
        Input:
            freq (float) : Frequency in Hz
        Output:
            None
        '''
        self.v2_stub.Sweep( SuperV2.SweepInput(frequency = frequency, power = self._power, offset = offset, method = method) )

    def do_get_power(self, query=True):
        '''
        Get output power of device
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            (float) microwave power in dBm
        '''
        return self.hfa_stub.GetAttenuation( HFA.EndpointIndex(value = 0) ).value
    def do_set_power(self, power = 0):
        '''
        Sets 0 to 30 dB RF attenuation.
        Output level depends on modulator input or pilot power
        (usually around -10 to 0 dBm max.)
        Input:
            power (float) : "Power" in dBm
        Output:
            None
        '''
        self._power=power
        self.hfa_stub.SetAttenuation( HFA.Double(index = HFA.EndpointIndex(value = 0), value = abs(power)) )
    
    def do_get_connection(self):
        '''
        Is Instrument connected to IPESH?
        '''
        return self._connection
    def do_set_connection(self, connect):
        '''
        True: (Re)establish connection to IPESH channel and init stubs
        False: Close channel
        '''
        if connect:
            try:
                self._channel.close()
            except:
                pass
            self._channel = grpc.insecure_channel(address)
            self.hfl_stub = HFL_grpc.HfLOServiceStub(self._channel)
            self.hfd_stub = HFD_grpc.HfDemodulatorServiceStub(self._channel)
            self.hfa_stub = HFA_grpc.HfAttenuatorServiceStub(self._channel)
            self.v2_stub = SuperV2_grpc.HfSuperV2ServiceStub(self._channel)
        else:
            try:
                self._channel.close()
            except:
                pass

    def do_get_status(self, query=True):
        '''
        Get status of output channel
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            True (on) or False (off)
        '''
        #self.get_all(query)
        # TODO: implement read
        return self._status
    def do_set_status(self, status):
        '''
        Set status of output (Bool)
        Legacy functionality, manually starting up LOs as required preferred to reduce risk of overheating
        Input:
            status (bool): enable LOs or disable everything
        Output:
            None
        '''
        self._status = status
        if status:
            self.on()
        else:
            self.off()

    # Detail
    def do_get_iflo_frequency(self):
        return self.hfl_stub.GetFrequency( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) ).value
    def do_set_iflo_frequency(self, value):
        self.hfl_stub.SetFrequency( HFL.WriteDouble(index=HFL.EndpointIndex(value=self.EPDict.index("IFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        self.do_set_demod_lo_match(int(value))

    def do_get_rflo_frequency(self):
        return self.hfl_stub.GetFrequency(HFL.EndpointIndex(value=self.EPDict.index("RFLO"))).value
    def do_set_rflo_frequency(self, value):
        self.hfl_stub.SetFrequency( HFL.WriteDouble(index=HFL.EndpointIndex(value=self.EPDict.index("RFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )

    def do_get_txpl_frequency(self):
        return self.hfl_stub.GetFrequency(HFL.EndpointIndex(value=self.EPDict.index("TXPL"))).value
    def do_set_txpl_frequency(self, value):
        self.hfl_stub.SetFrequency( HFL.WriteDouble(index=HFL.EndpointIndex(value=self.EPDict.index("TXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )

    def do_get_rxpl_frequency(self):
        return self.hfl_stub.GetFrequency(HFL.EndpointIndex(value=self.EPDict.index("RXPL"))).value
    def do_set_rxpl_frequency(self, value):
        self.hfl_stub.SetFrequency( HFL.WriteDouble(index=HFL.EndpointIndex(value=self.EPDict.index("RXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )

    def do_get_demod_lo_match(self):
        return self.hfd_stub.GetLoMatchMHz(HFD.EndpointIndex(value=0)).value
    def do_set_demod_lo_match(self, value):
        self.hfd_stub.SetLoMatchMHz(HFD.WriteNum(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_gain(self):
        return self.hfd_stub.GetGain(HFD.EndpointIndex(value=0)).value
    def do_set_demod_gain(self, value):
        self.hfd_stub.SetGain(HFD.WriteNum(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_gain_error(self):
        return self.hfd_stub.GetGainErrorCdB(HFD.EndpointIndex(value=0)).value
    def do_set_demod_gain_error(self, value):
        self.hfd_stub.SetGainErrorCdB(HFD.WriteDouble(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_attenuation(self):
        return self.hfd_stub.GetAttenuation(HFD.EndpointIndex(value=0))
    def do_set_demod_attenuation(self, value):
        self.hfd_stub.SetAttenuation(HFD.WriteNum(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_phase_error(self):
        return self.hfd_stub.GetPhaseErrorDeg(HFD.EndpointIndex(value=0)).value
    def do_set_demod_phase_error(self, value):
        self.hfd_stub.SetPhaseErrorDeg(HFD.WriteDouble(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_dc_offset_i(self):
        return self.hfd_stub.GetDcOffsetImV(HFD.EndpointIndex(value=0)).value
    def do_set_demod_dc_offset_i(self, value):
        self.hfd_stub.SetDcOffsetImV(HFD.WriteDouble(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))

    def do_get_demod_dc_offset_q(self):
        return self.hfd_stub.GetDcOffsetQmV(HFD.EndpointIndex(value=0)).value
    def do_set_demod_dc_offset_q(self, value):
        self.hfd_stub.SetDcOffsetQmV(HFD.WriteDouble(index=HFD.EndpointIndex(value=0), value=value))
        self.hfd_stub.UpdateWrite(HFD.EndpointIndex(value=0))
    
    # --- Legacy support --- 
    # Direct usage of proto stubs is preferrable and provides more detailed functionality
    def on(self):
        """
        Set power to original value and turn up all LOs at full power. 
        Obsolete function, manually starting up LOs as required preferred to reduce risk of overheating
        """
        self.do_set_power(self._power)
        self.v2_stub.Startup(SuperV2.StartInfo( lmx_list = [#SuperV2.LMXStart(name="RXPL", power_a=0, power_b=0), 
                                                            #SuperV2.LMXStart(name="TXPL", power_a=0, power_b=0), 
                                                            SuperV2.LMXStart(name="RFLO", power_a=0, power_b=0), 
                                                            SuperV2.LMXStart(name="IFLO", power_a=0, power_b=0)] ))
        self.v2_stub.UpdateWrite(SuperV2.EndpointList(devices=["TXPL", "RXPL", "IFLO", "RFLO"]))
    def off(self):
        """
        Set power as low as possible and disable all LOs
        """
        self.do_set_power(30)
        self.v2_stub.Startup(SuperV2.StartInfo( lmx_list = [SuperV2.LMXStart(name="RXPL"), 
                                                            SuperV2.LMXStart(name="TXPL"), 
                                                            SuperV2.LMXStart(name="RFLO"), 
                                                            SuperV2.LMXStart(name="IFLO")] ))
        self.v2_stub.UpdateWrite(SuperV2.EndpointList(devices=["TXPL", "RXPL", "IFLO", "RFLO"]))

    def set_rflo_powerdown(self, boolean):
        if boolean:
            self.hfl_stub.PowerDown( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        else:
            self.hfl_stub.PowerUp( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
    def set_iflo_powerdown(self, boolean):
        if boolean:
            self.hfl_stub.PowerDown( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        else:
            self.hfl_stub.PowerUp( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
    def set_txpilot_powerdown(self, boolean):
        if boolean:
            self.hfl_stub.PowerDown( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        else:
            self.hfl_stub.PowerUp( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
    def set_rxpilot_powerdown(self, boolean):
        if boolean:
            self.hfl_stub.PowerDown( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        else:
            self.hfl_stub.PowerUp( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )

    def set_rflo_enable_a(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputA( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        else:
            self.hfl_stub.DisableOutputA( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
    def set_iflo_enable_a(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputA( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        else:
            self.hfl_stub.DisableOutputA( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
    def set_txpilot_enable_a(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputA( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        else:
            self.hfl_stub.DisableOutputA( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
    def set_rxpilot_enable_a(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputA( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        else:
            self.hfl_stub.DisableOutputA( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )

    def set_rflo_enable_b(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputB( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        else:
            self.hfl_stub.DisableOutputB( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
    def set_iflo_enable_b(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputB( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        else:
            self.hfl_stub.DisableOutputB( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
    def set_txpilot_enable_b(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputB( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        else:
            self.hfl_stub.DisableOutputB( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
    def set_rxpilot_enable_b(self, boolean):
        if boolean:
            self.hfl_stub.EnableOutputB( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        else:
            self.hfl_stub.DisableOutputB( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )

    def set_rflo_output_a_power(self, value):
        self.hfl_stub.SetOutputPowerA( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("RFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
    def set_iflo_output_a_power(self, value):
        self.hfl_stub.SetOutputPowerA( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("IFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
    def set_txpilot_output_a_power(self, value):
        self.hfl_stub.SetOutputPowerA( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("TXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
    def set_rxpilot_output_a_power(self, value):
        self.hfl_stub.SetOutputPowerA( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("RXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )

    def set_rflo_output_b_power(self, value):
        self.hfl_stub.SetOutputPowerB( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("RFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RFLO")) )
    def set_iflo_output_b_power(self, value):
        self.hfl_stub.SetOutputPowerB( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("IFLO")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("IFLO")) )
    def set_txpilot_output_b_power(self, value):
        self.hfl_stub.SetOutputPowerB( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("TXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("TXPL")) )
    def set_rxpilot_output_b_power(self, value):
        self.hfl_stub.SetOutputPowerB( HFL.WriteNum(index=HFL.EndpointIndex(value=self.EPDict.index("RXPL")), value=value) )
        self.hfl_stub.UpdateWrite( HFL.EndpointIndex(value=self.EPDict.index("RXPL")) )
    """
    def set_linopt(self, value):
        self._rpc.setlinopt(value)
    """
    def set_hf_attenuation(self, value):
        self.do_set_power(value)
    def reinitialise(self):
        self.v2_stub.Reset(SuperV2.EndpointList(devices=["TXPL", "RXPL", "IFLO", "RFLO", "Demod"]))    

    def set_rxpilot_frequency(self, value):
        self.do_set_rxpl_frequency(value)
    def set_txpilot_frequency(self, value):
        self.do_set_txpl_frequency(value)
    
    # These legacy functions should now be generated from Qkit parameters automatically
    # Warning: Units changed for gain/phase error. LO freqs can now be floats
    """
    def set_demod_attenuation(self, value):
        self._nominalrxatt = value
        self._rpc.writeval("Demodulator", "attenuation", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
    def set_demod_gain(self, value):
        self._rpc.writeval("Demodulator", "gain", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
    def set_demod_gain_error(self, value):
        self._rpc.writeval("Demodulator", "gain_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
    def set_demod_phase_error(self, value):
        self._rpc.writeval("Demodulator", "phase_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)

    def set_rflo_frequency(self, value):
        self._rpc.writeval("RFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("RFLO", "recalibrate", 1)
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)
        self._rffreq = value/1e9
    def set_iflo_frequency(self, value):
        self._rpc.writeval("ZFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("ZFLO", "recalibrate", 1)
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)
        self._rpc.writeval("Demodulator", "lo_match_mhz", int(value/1e6))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        self._iffreq = value/1e9
    """



# progress bar handler by AS/JB@KIT 09/2014, 08/2015
# jupyter adjust HR@KIT/2016
# use:
# import Progress_Bar
# initialize and create empty progress bar by instantiating, passing the number of iterations: e.g. p = Progress_Bar.Progress_Bar(100)
# to iterate the progress bar call p.iterate()
# pass _dummy=True to initialize a _dummy progressbar if you don't want to see it.



import time,sys

debug = False
try_legacy = False

try:
    pb_list
    if debug:
        print("List of progressbars with %i items found"%(len(pb_list)))
except NameError:
    pb_list = []

def hourformat(time):
    time = int(time)
    return "%i:%02i:%02i"%(time/60/60,time/60%60,time%60)
try:
    "try using the jupyter progress bar"
    #import somemodule.that.doesnt.exist

    from ipywidgets import IntProgress, HTML
    from IPython.display import display
    class Progress_Bar(object):
        def __init__(self,max_it,name= 'Progress:',est_cycle_time=None,dummy=False):
            if debug:
                print("new style progress bar")
            self._dummy=dummy
            if self._dummy:
                return
            self.starttime = time.time()
            self.start_eta_time = self.starttime
            
            self.max_it = max_it
            self.name = name
            self.progr = 0
            
            #check for old (finished) progressbar with the same name
            for p in pb_list:
                if p[0] == self.name:
                    p[1].close()
                    p[2].close()

            self.pb = IntProgress(
                value=0,
                min=0,
                max=self.max_it,
                description=self.name,
                layout={"width": "95%"},
                )     
            
            self.pi = HTML(
                #value = "(0/%i) <br>&#9992; -?-    <br>&#128336;  --:--:--   (estimated)<br>&#10010;  00:00:00 (elapsed) <br>&#9866; --:--:--  (remaining)"% (self.max_it),
                value = "<table style='width:100%%'><tr><td>%s (%i/%i) </td><td>&#9992; %s    </td><td>&#128336;  %s   (estimated)</td><td>&#10010;  %s (elapsed) </td><td>&#9866;  %s (remaining)</td></tr></table>"%("",
                    0,
                    self.max_it,
                    "-?-" if est_cycle_time==None else time.strftime('%Y-%m-%d (%a) %H:%M:%S', time.localtime(time.time() + est_cycle_time * self.max_it )),
                    "--:--:--" if est_cycle_time==None else hourformat(est_cycle_time*self.max_it),
                    "00:00:00",
                    "--:--:--" if est_cycle_time==None else hourformat(est_cycle_time*self.max_it)),
                )

            display(self.pi)
            display(self.pb)            
            
        def reset(self,max_it=None,name=None):
            if self._dummy:
                return
            if max_it is not None:
                self.max_it = max_it
            if name is not None:
                self.name = name
                self.pb.description = self.name
            self.progr = 0
            self.starttime = time.time()
            self.start_eta_time = self.starttime
            self.pb.color = None
            self.pb.bar_style = ""
            self.pb.max = max_it
            self._update()
                    
        def set(self,iteration,param=""):
            if self._dummy:
                return
            self.progr = iteration
            self._update(param)
        
        def iterate(self, param="", addend=1):
            if self._dummy:
                return
            self.progr += addend
            self._update(param)
        
        def _update(self,param=""):
            if self._dummy:
                return
            progr_info = "<table style='width:100%%'><tr><td>%s (%i/%i) </td><td>&#9992; %s    </td><td>&#128336;  %s   (estimated)</td><td>&#10010;  %s (elapsed) </td><td>&#9866;  %s (remaining)</td></tr></table>"%(param,     #"%s (%i/%i) &#10148;  ETA: %s &#10148; Time elapsed: %s" %(param,
                    self.progr,
                    self.max_it,
                    time.strftime('%Y-%m-%d (%a) %H:%M:%S', time.localtime(time.time() + float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it  - self.progr))), #ETA
                    hourformat(self.start_eta_time-self.starttime+float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it -(0 if self.progr == 1 else 1) )), #estimated
                    hourformat(time.time()-self.starttime), #elapsed
                    hourformat(float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it  - self.progr))) #remaining
            self.pi.value = progr_info
            self.pb.value = self.progr
            if self.progr == 1:
                "this is a little academic, but the time between the first and the second iteration has usually a time lag."
                self.start_eta_time = time.time()
                            
            if self.progr == self.max_it: #last iteration
                self.pb.color = "green"
                self.pb.bar_style = "success"
                pb_list.append([self.name,self.pb,self.pi]) #append to the list of done PBs
                #progr_info = "%s (%i/%i) &#9992; %s    &#10010;  %s  "%(param,     #"%s (%i/%i) &#10148;  ETA: %s &#10148; Time elapsed: %s" %(param,
                #    self.progr,
                #    self.max_it,
                #    time.ctime(time.time() + float(time.time()-self.start_eta_time)/self.progr * (self.max_it - 1  - self.progr)),
                #    #time.strftime('%H:%M:%S', time.gmtime(self.start_eta_time-self.starttime+float(time.time()-self.start_eta_time)/self.progr * (self.max_it - 1 ))),
                #    time.strftime('%H:%M:%S', time.gmtime(time.time()-self.starttime)))
                #    #time.strftime('%H:%M:%S', time.gmtime(float(time.time()-self.start_eta_time)/self.progr * (self.max_it - 1  - self.progr))))
                #self.pi.value = progr_info
            
        def abort(self):
            if self._dummy:
                return
            self.pb.color = "red"
            self.pb.bar_style = "danger"
            pb_list.append([self.name, self.pb, self.pi])  # append to the list of done PBs
            
                    
except ImportError as e:
    if debug:
        print(e)
    "Most likely not yet in the jupyter environment ... Falling back to old style progress bar, untouched."
    import uuid
    from IPython.display import HTML, Javascript, display
    from IPython.display import clear_output


    class Progress_Bar(object):

        def __init__(self, max_it, name = 'Progress:',est_cycle_time=None,dummy=False):
            if debug:
                print("old style progress bar")
            #create HTML progress bar
            self._dummy = dummy
            if self._dummy:
                return
            self.divid = str(uuid.uuid4())
            self.max_it = max_it
            self.name = name
            self.progr = 0

            #delete existing progress bar with same name
            display(Javascript("if(document.getElementById('%st') !== null) document.getElementById('%st').parentNode.remove()"%(self.name,self.name)))
            #display(Javascript("if(document.getElementById('%s') !== null) document.getElementById('%s').remove()"%(self.name,self.name)))
               
            outp = "<table style='width:100%%;border:none'><tr style='border:none'><td style='border:none'>%s (%i/%i) </td><td style='border:none'>&#9992; %s    </td><td style='border:none'>&#128336;  %s   (estimated)</td><td style='border:none'>&#10010;  %s (elapsed) </td><td style='border:none'>&#9866;  %s (remaining)</td></tr></table>"%("",
                    0,
                    self.max_it,
                    "-?-" if est_cycle_time==None else time.strftime('%Y-%m-%d (%a) %H:%M:%S', time.localtime(time.time() + est_cycle_time * self.max_it )),
                    "--:--:--" if est_cycle_time==None else hourformat(est_cycle_time*self.max_it),
                    "00:00:00",
                    "--:--:--" if est_cycle_time==None else hourformat(est_cycle_time*self.max_it))
           
            self.pb = HTML(
            """
            <div id="%s_title"> %s</div>
            <div id="%s1" style="border: 1px solid black; width:900px">
              <div id="%s0" style="text-align: center; color:white; background-color:blue; width:0%%">&nbsp;</div>
            </div>
            <div id="%s_text">%s</div>
            """ % (self.divid,self.name,self.divid,self.divid,self.divid,outp))
            display(self.pb)
            display(Javascript("$('div#%s').width('%i%%')" % (self.divid, 100*self.progr/self.max_it)))
            sys.stdout.flush()
            self.starttime = time.time()
            self.start_eta_time = time.time()

        def iterate(self,param=""):
            if self._dummy:
                return
            self.progr += 1
            self._update(param)
            
        def _update(self,param=""):
            if self._dummy:
                return
            display(Javascript("$('div#%s0').width('%i%%');" % (self.divid, 100*self.progr/self.max_it)))
            outp = "<table style='width:100%%;border:none'><tr style='border:none'><td style='border:none'>%s (%i/%i) </td><td style='border:none'>&#9992; %s    </td><td style='border:none'>&#128336;  %s   (estimated)</td><td style='border:none'>&#10010;  %s (elapsed) </td><td style='border:none'>&#9866;  %s (remaining)</td></tr></table>"%(param,     #"%s (%i/%i) &#10148;  ETA: %s &#10148; Time elapsed: %s" %(param,
                    self.progr,
                    self.max_it,
                    time.strftime('%Y-%m-%d (%a) %H:%M:%S', time.localtime(time.time() + float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it  - self.progr))), #ETA
                    hourformat(self.start_eta_time-self.starttime+float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it )), #estimated
                    hourformat(time.time()-self.starttime), #elapsed
                    hourformat(float(time.time()-self.start_eta_time)/(self.progr-(0 if self.progr == 1 else 1)) * (self.max_it  - self.progr))) #remaining
            if self.progr == 1:
                "this is a little academic, but the time between the first and the second iteration has usually a time lag."
                self.start_eta_time = time.time()
            #outp = "%s (%i/%i) &#10148;  ETA: %s &#10148; Time elapsed: %s"%(param,self.progr,self.max_it,time.ctime(time.time() + float(time.time()-self.starttime)/self.progr * (self.max_it - self.progr)),time.strftime('%H:%M:%S',time.gmtime(time.time()-self.starttime)))
            display(Javascript("document.getElementById('%s_text').innerHTML = \"%s\";"%(self.divid,outp)))

            if self.progr == self.max_it:   #end of progress bar
                #Turn the status bar into green
                #Delete all <div> containers
                outp = "%s (%i/%i) &#9992; %s    &#10010;  %s  "%(param,     #"%s (%i/%i) &#10148;  ETA: %s &#10148; Time elapsed: %s" %(param,
                    self.progr,
                    self.max_it,
                    time.strftime('%Y-%m-%d (%a) %H:%M:%S'),
                    hourformat(time.time()-self.starttime))
                display(Javascript("document.getElementById('%s_title').parentNode.remove();"%self.divid)) #title
                self.pb = HTML(
                """
                <div id="%st"> %s</div>
                <div id="%s" style="border: 1px solid black; width:900px;text-align: center; color:white; background-color:green;">%s
                </div>
                """ % (self.name, self.name, self.name, outp))
                display(self.pb)
        
        def reset(self,max_it=None,name=None):
            if self._dummy:
                return
            if max_it is not None:
                self.max_it = max_it
            if name is not None:
                self.name = name
                self.pb.description = self.name
            self.progr = 0
            self.starttime = time.time()
            self.start_eta_time = self.starttime
            self._update()
                    
        def set(self,iteration,param=""):
            if self._dummy:
                return
            self.progr = iteration
            self._update(param)
            
        def abort(self):
            if self._dummy:
                return
            display(Javascript("$('div#%s0').css('background-color', 'red');" % (self.divid)))


import logging
from pathlib import Path
import platform
from types import ModuleType
from typing import Callable
import shutil
import sys

try:
    from importlib.resources import files, as_file
    from importlib.abc import Traversable
except ModuleNotFoundError:
    from importlib_resources import files, as_file
    from importlib.abc import Traversable

# Install Logging, Minimal Formatting
logging.basicConfig(level=logging.DEBUG, 
                    format='[%(levelname).1s] %(message)s')
logger = logging.getLogger(__name__)

# Decorator for marking a install step as optional
def optional(desc: str) -> Callable[[Callable], Callable]:
    import functools
    def wrapper(f: Callable) -> Callable:
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            run = input(f"Run optional task '{f.__name__}'? ({desc}) (y/n): ").lower() == 'y'
            if run:
                return f(*args, **kwargs)
            return f
        return wrapped
    return wrapper

# Decorator for requiring admin privileges
def windows_admin_required(f: Callable) -> Callable:
    import ctypes
    import functools

    if not ctypes.windll.shell32.IsUserAnAdmin():
        @functools.wraps(f)
        def warning(*args, **kwargs):
            logger.warning("Admin privileges required for this task. Skipping.")
        return warning
    else:
        return f


# Get binary path for installed command:
def get_binary(name: str) -> Path:
    import shutil
    return str(Path(shutil.which(name, path=Path(sys.argv[0]).parent)).absolute())


def copy_named_template(source_cache: Traversable, target_path: Path, target_name: str, human_readable: str | None = None):
    package_file: Traversable = source_cache / f'{target_name}-tpl'
    logging.debug("Got reference %s: %s", human_readable if human_readable is not None else target_name, package_file)
    target = (target_path / target_name)
    if target.exists():
        logging.warning("%s already exists, skipping.", target)
        return
    with as_file(package_file) as f:
        shutil.copyfile(f, target)

#################################
# Define Universal Scripts Here #
#################################

def create_base_structure(pwd: Path):
    """
    Check if the PWD exists and create the folders 'notebooks', 'data' and 'logs'.
    """
    assert pwd.exists(), "The current working directory must exist!"

    # Ensure Required Directories
    notebook_dir = pwd / "notebooks"
    notebook_dir.mkdir(exist_ok=True)

    data_dir = pwd / "data"
    data_dir.mkdir(exist_ok=True)
    
    log_dir = pwd / "logs"
    log_dir.mkdir(exist_ok=True)

    # Copy Files From Module
    from qkit import install
    install_files = files(install)

    copy_named_template(install_files, pwd, "jupyter_lab_config.py", "Jupyter Lab Config")
    copy_named_template(install_files, pwd, "qkit_local_config.py", "Qkit Local Config")



UNIVERSAL_SCRIPTS: list[Callable[[Path], None]] = [create_base_structure]

##############################
# Define System Scripts Here #
##############################

def windows_install_scripts(pwd: Path):
    from qkit.install import windows
    windows_script_files = files(windows)

    copy_named_template(windows_script_files, pwd, "launch.bat", "Launch Jupyter Lab Script")

@windows_admin_required
@optional("Associate .h5 files with Qviewkit. Modifies the Registry.")
def windows_associate_h5(pwd: Path):
    import winreg
    # Create file type if it does not exist.
    try:
        base_key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, r"qviewkit.h5")
        (value, type) = winreg.QueryValueEx(base_key, None)
        logging.info("File Type qviewkit.h5 already exists.")
        command_key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, r"qviewkit.h5\shell\open\command")
        (value, type) = winreg.QueryValueEx(command_key, None)
        logging.info("Command already exists.")
    except FileNotFoundError:
        logging.info("Creating file type qviewkit.h5...")
        base_key = winreg.CreateKeyEx(winreg.HKEY_CLASSES_ROOT, r"qviewkit.h5")
        winreg.SetValue(base_key, None, winreg.REG_SZ, "QViewKit HDF5 File")
        winreg.CloseKey(base_key)

        open_command_key = winreg.CreateKey(winreg.HKEY_CLASSES_ROOT, r"qviewkit.h5\shell\open\command")
        winreg.SetValue(open_command_key, None, winreg.REG_SZ, f'"{get_binary("qviewkit")}" -f "%1"')
        winreg.CloseKey(open_command_key)
        logging.info("Qkit registered.")


    # Create Association if it does not exist.
    try:
        key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, r".h5")
        (value, type) = winreg.QueryValueEx(key, None)
        logging.info("Association of .h5 to %s already exists.", value)
    except FileNotFoundError:
        assoc_key = winreg.CreateKey(winreg.HKEY_CLASSES_ROOT, r".h5")
        winreg.SetValue(assoc_key, None, winreg.REG_SZ, "qviewkit.h5")
        winreg.CloseKey(assoc_key)
        logging.info("Association created.")

@windows_admin_required
@optional("Disable Smart Sorting in Windows Explorer. Edits the registry.")
def windows_disable_smart_sorting(pwd: Path):
    import winreg
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows\CurrentVersion\Policies\Explorer")
    winreg.SetValue(key, "NoStrCmpLogical", winreg.REG_DWORD, 1)

def windows_set_config_path(pwd: Path):
    import os
    config_path = pwd / "qkit_local_config.py"
    os.system(f'setx QKIT_LOCAL_CONFIG "{str(config_path.absolute())}"')

def linux_set_env(pwd: Path):
    import os
    if 'QKIT_LOCAL_CONFIG' in os.environ or 'QKIT_VENV' in os.environ:
        logger.warning("Qkit environment variables already set! Is qkit already installed?")
        return
    
    # Construct export command for local config and venv
    config_path = pwd / "qkit_local_config.py"
    venv_activate = Path(sys.executable).parent / "activate"
    var_export = f'\n#QKIT INSTALL\nQKIT_LOCAL_CONFIG="{str(config_path.absolute())}"\nQKIT_VENV="{str(venv_activate)}"\n\n'

    # Install Config for local user
    with open(Path.home() / ".profile", mode="a") as f:
        f.write(var_export)

def linux_install_desktop_files(pwd: Path):
    def install_file(fname):
        import os
        from qkit.install import linux
        linux_files = files(linux)
        tpl_file: Traversable = linux_files / fname
        desktop_file_target = Path.home() / '.local' / 'share' / 'applications'
        with as_file(tpl_file) as f:
            os.system(f"desktop-file-install --dir='{str(desktop_file_target)}' '{str(f)}'")
    
    install_file('qkit-qviewkit.desktop')
    install_file('qkit-qviewkit-url.desktop')
    install_file('qkit-jupyter-lab.desktop')

SYSTEM_SCRIPTS: dict[str, list[Callable[[Path], None]]] = {
    "Windows": [windows_install_scripts, windows_associate_h5, windows_set_config_path, windows_disable_smart_sorting],
    "Linux": [linux_set_env, linux_install_desktop_files],
    "Darwin": []
}

# Main Install Routine
def main():
    """
    Get and List Relevant paths (for diagnosis, should this fail) and setup for qkit usage.
    """
    logger.info("Qkit Setup")
    logger.info("Running from %s", sys.argv[0])

    system = platform.system()
    logger.info("Operating System Type: %s", system)
    if system not in SYSTEM_SCRIPTS.keys():
        logger.error("Unsupported Operating System.")
        return

    pwd = Path.cwd()
    logger.info("User Install Path: %s", pwd)

    logger.info("Looking for qkit...")
    try:
        import qkit
        logger.info("Qkit installed, path: %s", qkit.__file__)
    except ImportError:
        logger.error("Qkit not importable, can't install.")
        return
    
    scripts = UNIVERSAL_SCRIPTS + SYSTEM_SCRIPTS[system]
    logger.info("Running %s install tasks...", len(scripts))
    for script in scripts:
        logger.info("Running task '%s'...", script.__name__)
        try:
            script(pwd)
        except:
            logger.exception("Setup step '%s' failed with the following Exception:", script.__name__)

    logging.info("Done.")

if __name__ == "__main__":
    main()


import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as si

class CurveXtractor():
    """
    Class for graphical curve extraction.
    Use the mouse to find points in your data.
    Use the keyboard to generate a spline following your points.
    Finally, extract a curve from the data, which is in the vicinity of the spline.
    Note, use of the module requires the widget backend for matplotlib.
    
    Attributes:
    
    set_data: set x-, y- and zdata from which curves will be extracted.
    set_smoothness: set smoothness of the splines used to approximate the curves in the data.
    set_points: manually add points and curves you want to extract from the data.
    extract_points: extract curves from the data, which are follow the splines.
    plot_results: plot extracted points.
    
    """
    def __init__(self, xdata = [0], ydata = [0], zdata = [0], smoothness = 0.1):
        """
        Inits CurveXtractor with xdata, ydata, zdata and smoothness of the splines.
        
        Args:
            xdata:      xaxis data
            ydata:      yaxis data
            zdata:      2d matix of zdata
            smoothness: smoothness of the spline (can also be changed later on in the procedure).
        """
        if plt.get_backend().find("nbagg") == -1:
            print("Module requires using the widget backend for matplotlib.\nActivate with %matplotlib widget.")
            return
        self.xvals = 0 
        self.yvals = 0 
        self.zvals = 0
        self._smoothness = smoothness
        
        self._status = True
        self._infotxt = 0
        
        self._points_x = []
        self._points_y = []
        self._curve_x = []
        self._curve_y = []
        self._x_results = []
        self._y_results = []
        
        self._col = ["w", "C1", "C3", "C6", "C8"]
        self._pointplots = [[]]
        self._curveplots = []
        self._result_plots = []
        self._xlim = (min(xdata), max(xdata))
        self._ylim = (min(ydata), max(ydata))
        
        self._splines = []


    def set_data(self, xdata, ydata, zdata):
        """
        Adds x-, y- and zdata.
        
        Args:
            xdata: xaxis data
            ydata: yaxis data
            zdata: 2d matix of zdata
        """
        self.xvals = xdata
        self.yvals = ydata
        self.zvals = zdata
        self._xlim = (min(xdata), max(xdata))
        self._ylim = (min(ydata), max(ydata))
        return True


    def set_smoothness(self, smoothness):
        """
        Set smoothness of the splines.
        
        Args:
            smoothness: Smoothness parameter. Small smoothness forces the spline closer to the points used to generate it.
                        Higher smoothness allows the spline to deviate more strongly from the given curve.
        
        """
        self._smoothness = smoothness
        return True


    def _click_event(self, event):
        """
        button_press_event:
            left_mouse:  add a point (mouse position) to the current curve.
            right_mouse: unify previously added points to a new curve and initialize a new curve.
        """
        if self._infotxt != 0:
            self._removeinfo()
            return True
        if self._status:
            if event.button == 1:
                # add point to plot on left mouse click
                if (event.xdata < self._xlim[-1]) and (event.xdata > self._xlim[0]) and (event.ydata < self._ylim[-1]) and (event.ydata > self._ylim[0]):
                    self._addpoint(event.xdata, event.ydata)
            elif (event.button == 3) and (len(self._points_x) > 1):
                # group current points as new curve on right mouse click
                self._addcurve()
        return


    def _key_event(self, event):
        """
        key_press_event:
            enter:  generate splines from previously added points. Or revert to point adding mode.
            delete: remove latest point or curve from data and plot.
        """
        if self._infotxt != 0:
            self._removeinfo()
            return True
        if (event.key == "delete") and self._status:
            self._del_last()
        elif event.key == "enter":
            if len(self._curveplots) > 0:
                for plot in self._curveplots:
                    plot.pop(0).remove()
                self._curveplots = []
                self._splines = []
                if self._status:
                    self._generate_splines()
                else:
                    for i in range(0, len(self._curve_x)):
                        temp_plot = plt.plot(np.array(self._curve_x[i])[np.argsort(self._curve_x[i])],
                                             np.array(self._curve_y[i])[np.argsort(self._curve_x[i])],
                                             self._col[i % len(self._col)] + ".--")
                        self._curveplots.append(temp_plot)
                plt.xlim(self._xlim)
                plt.ylim(self._ylim)
                self._status = not self._status
        return


    def set_points(self, show_info = True):
        """
        This routine starts the graphical interface for adding points.
        
        Use left mouse button to add points to the current curve.
        Use right mouse button to add another curve.
        Press delete (entf) to remove the last point.
        
        Once you are finished press enter to calculate splines following your curves.
        If you are satisfied with the result, continue with extract_points.
        In case you want to change something, press enter to return to the editing mode.
        
        Args:
            show_info: show the information of the doc string in the plot.
        """
        self._fig = plt.figure(figsize = (9, 6))
        plt.pcolormesh(self.xvals, self.yvals, self.zvals)
        info = ("""Use left mouse button to add points to the current curve.\nUse right mouse button to add another curve.\nPress delete (entf) to remove the last point.\n
        
        Once you are finished press enter to calculate splines following your curves.
        If you are satisfied with the result, continue with extract_points.
        In case you want to change something, press enter to return to the editing mode.""")

        self._cid = self._fig.canvas.mpl_connect("button_press_event", self._click_event)
        self._cid = self._fig.canvas.mpl_connect("key_press_event", self._key_event)
        if show_info:
            self._showinfo(info)
        return True


    def _addpoint(self, x, y):
        """
        Add point (x,y) to plot and (points_x, points_y).
        
        Args:
            x: x coord of the point.
            y: y coord of the point.
        """
        temp_plot = plt.plot([x], [y], self._col[len(self._curve_x)%len(self._col)] + ".")
        self._pointplots[-1].append(temp_plot)
        self._points_x.append(x)
        self._points_y.append(y)
        return True


    def _del_last(self):
        """
        Remove last point or curve from plot and data.
        """
        if (len(self._pointplots[-1]) == 0) and (len(self._pointplots) > 1):
            del self._pointplots[-1]
        if len(self._points_x) > 0:
            self._pointplots[-1][-1].pop(0).remove()
            del self._pointplots[-1][-1]
            self._points_x = self._points_x[:-1]
            self._points_y = self._points_y[:-1]
        elif len(self._curve_x) > 0:
            self._points_x = self._curve_x[-1]
            self._points_y = self._curve_y[-1]
            self._curve_x = self._curve_x[:-1]
            self._curve_y = self._curve_y[:-1]
            self._curveplots[-1].pop(0).remove()
            self._curveplots = self._curveplots[:-1]
        return True


    def _addcurve(self):
        """
        Unify previously added points to a curve.
        Start a new curve.
        """
        temp_plot = plt.plot(np.array(self._points_x)[np.argsort(self._points_x)],
                             np.array(self._points_y)[np.argsort(self._points_x)],
                             self._col[len(self._curve_x)%len(self._col)] + ".--")
        self._curveplots.append(temp_plot)
        self._curve_x.append(self._points_x)
        self._curve_y.append(self._points_y)
        self._pointplots.append([])
        self._points_x = []
        self._points_y = []
        return True


    def _generate_splines(self):
        """
        For each curve, calculate a function (spline), following the points which were added to the curve.
        If there are only 3 points given, a parabola is calculated.
        For 2 points, a straight line is used.
        """
        warning = ""
        for i in range(0, len(self._curve_x)):
            if len(self._curve_x[i]) == 3:
                # if only 3 points are given, calculate exact parabola
                warning += "In curve " + str(i) + ", only 3 points are given.\n" + "A parabola is used instead of a spline.\n\n"
                g = (self._curve_y[i][0] - self._curve_y[i][1])/(self._curve_y[i][2] - self._curve_y[i][1])
                x0 = 0.5 * ((self._curve_x[i][0]**2 - self._curve_x[i][1]**2 * (1 - g) - g*self._curve_x[i][2]**2)/
                            (self._curve_x[i][0] - self._curve_x[i][1] - g*(self._curve_x[i][2] - self._curve_x[i][1])))
                m = (self._curve_y[i][1] - self._curve_y[i][0])/((self._curve_x[i][1] - x0)**2 - (self._curve_x[i][0] - x0)**2)
                c = self._curve_y[i][0] - m*(self._curve_x[i][0] - x0)**2
                fct = lambda x: m * (x - x0)**2 + c
            elif len(self._curve_x[i]) == 2:
                # if only 2 points are given, a straight line is used
                warning += "In curve " + str(i) + ", only 2 points are given.\n" + "A straight line is used instead of a spline.\n\n"
                m = (self._curve_y[i][1] - self._curve_y[i][0])/(self._curve_x[i][1] - self._curve_x[i][0])
                c = self._curve_y[i][1] - m * self._curve_x[i][1]
                fct = lambda x: m * x + c
            else:
                # if there are more than 3 points given, an acutual spline is calculated.
                try:
                    fct = si.UnivariateSpline(np.array(self._curve_x[i])[np.argsort(self._curve_x[i])],
                                              np.array(self._curve_y[i])[np.argsort(self._curve_x[i])],
                                              s = self._smoothness)
                except ValueError:
                    warning += "Points in curve " + str(i) + " too dense.\n" + "Removing every 2nd point.\n\n"
                    for plot in self._pointplots[-len(self._curve_x[i]):][::2]:
                        plot.pop(0).remove()
                    self._pointplots = self._pointplots[::2]
                    self._curve_x[i] = self._curve_x[i][::2]
                    self._curve_y[i] = self._curve_y[i][::2]
                    try:
                        fct = si.UnivariateSpline(np.array(self._curve_x[i])[np.argsort(self._curve_x[i])],
                                                  np.array(self._curve_y[i])[np.argsort(self._curve_x[i])],
                                                  s = self._smoothness)
                    except:
                        warning += "Although points were removed from curve " + str(i) + " a spline could not be calculated.\n"
                        warning += "Please try again later.\n\n"
                        fct = lambda x: x * 0
            
            self._splines.append(fct)
            temp_plot = plt.plot(self.xvals, fct(self.xvals), self._col[i % len(self._col)] + "--")
            self._curveplots.append(temp_plot)
        if len(warning) > 0:
            self._showinfo(warning[:-2])
        return True


    def _showinfo(self, string):
        """
        Show a given string as text in the plot.
        
        Args:
            string: text to be shown in the plot.
        """
        self._infotxt = plt.text(np.mean(self._xlim), np.mean(self._ylim), string,
                                 horizontalalignment = "center", verticalalignment = "center",
                                 bbox=dict(boxstyle="square", ec=(1,0,0), fc=(1,1,1)))
        return True


    def _removeinfo(self):
        """
        Remove a text previously (added with _showinfo) from the plot.
        """
        self._infotxt.remove()
        self._infotxt = 0
        return True
    
    
    def find_along_splines(self, width = 0, peak = False):
        """
        Search the extremum along the spline functions, in a certain width.

        Args:
            width: Width (in units of y-axis) around which points are considered for the search.
                This can also be an array, each element corresponding to one curve.
                The default value for width is 5% of the y-range.
            peak:  If True/False, the maximum/minimum in the specified region is returned.
        """
        fct_num = len(self._splines)
        self._x_results = [[] for i in range(fct_num)]
        self._y_results = [[] for i in range(fct_num)]

        if width == 0:
            width = 0.05 * (np.amax(self.yvals) -np.amin(self.yvals))

        for i, x in enumerate(self.xvals):
            for j, fct in enumerate(self._splines):
                val = fct(x)
                ind_min = np.argmin(np.abs(self.yvals - (val - width/2.)))
                ind_max = np.argmin(np.abs(self.yvals - (val + width/2.)))
                temp_data = self.zvals[ind_min:ind_max, i]
                if len(temp_data) > 0:    
                    if peak:
                        y = self.yvals[ind_min + np.argmax(temp_data)]
                    else:
                        y = self.yvals[ind_min + np.argmin(temp_data)]
                    self._x_results[j].append(x)
                    self._y_results[j].append(y)
        return True


    def plot_results(self):
        """
        Plot results from find_along_splines.
        """
        for plot in self._result_plots:
            plot.pop(0).remove()
        self._result_plots = []
        if (len(self._x_results) > 0):
            for i in range(len(self._x_results)):
                temp_plot = plt.plot(self._x_results[i], self._y_results[i], self._col[i % len(self._col)])
                self._result_plots.append(temp_plot)
        return True


    def get_results(self):
        """
        Get results from find_along_splines.

        Returns:
            x_results: list of arrays containing x values (1 for each spline).
            y_results: list of arrays containing y values (1 for each spline).
        """
        return self._x_results, self._y_results


# -*- coding: utf-8 -*-
"""
@author: hannes.rotzinger@kit.edu / 2015,2016,2017 
         marco.pfirrmann@kit.edu / 2016, 2017
@license: GPL
"""

import sys,os
import qkit
# support both PyQt4 and 5
in_pyqt5 = False
in_pyqt4 = False
try:
    from PyQt5 import QtCore
    import PyQt5.QtWidgets as QtGui
    from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject,QTimer
    from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog
    in_pyqt5 = True
except ImportError as e:
    pass
if not in_pyqt5:    
    try:
        from PyQt4 import QtCore, QtGui
        from PyQt4.QtCore import *
        from PyQt4.QtGui import *
        in_pyqt4 = True
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)

import h5py
from qkit.gui.qviewkit.main_view import Ui_MainWindow
from qkit.core.lib.misc import  str3

class DatasetsWindow(QMainWindow, Ui_MainWindow):
    """DatasetsWindow fills the frame of the Ui_MainWindow.

    This class reads the file content and displays i.e. the dataset tree and
    the dataset attributes and handles the signal slots from the UI buttons
    i.e. the checking/unchecking of the live option.
    """
    refresh_signal = pyqtSignal()
    pw_refresh_signal = pyqtSignal()
    def __init__(self,DATA):
        """Inits DatasetsWindows with a DATA object coming from the main call
        in main.py.

        The Ui_MainWindow gets set up, the signal slots to the UI buttons get
        connected and set_cmd_options() evaluates the parsed arguments of the 
        main call.
        Once a file is opened the handle functions emmit signals as soon as 
        some user interaction happens (i.e. a dataset is clicked to be opened)
        and update_plots() emmits the update signal.
        """
        self.DATA = DATA
        #QMainWindow.__init__(self,None,Qt.WindowStaysOnTopHint)
        QMainWindow.__init__(self)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        
        # set up User Interface (widgets, layout...)
        self.setupUi(self)
        self.setWindowTitle("Qviewkit")
        self.treeWidget.setHeaderHidden(True)
        
        self.refreshTime_value = 2000
        self.tree_refresh  = True
        self._force_live_plot = False
        self._setup_signal_slots()        
        self.setup_timer()
        self.set_cmd_options()
        
        
    def setup_timer(self):
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_file)
        self.timer.timeout.connect(self.live_update_onoff)
            
    def set_cmd_options(self):
        # prepare datasets to display
        if self.DATA.args.datasets:
            dss = self.DATA.args.datasets.split(',')
            
            for ds in dss:
                fd = ds.split('/')
                if len(fd) == 1:
                    dsp = "/entry/data0/"+fd[0]
                    self.DATA.ds_cmd_open[dsp] = True
                if len(fd) == 2:
                    dsp = "/entry/"+fd[0]+"/"+fd[1]
                    self.DATA.ds_cmd_open[dsp] = True
            print("Display:", self.DATA.ds_cmd_open)
        if self.DATA.args.file:
            self.DATA.DataFilePath = self.DATA.args.file
            self.update_file()
        if self.DATA.args.refresh_time:
            self.refreshTime.setValue(self.DATA.args.refresh_time)
            self._refresh_time_handler(self.DATA.args.refresh_time)
        if self.DATA.args.live_plot:
            self.liveCheckBox.click()
            #self.live_update_onoff()
            
    def _setup_signal_slots(self):
        self.treeWidget.itemChanged.connect(self.handleChanged)
        self.treeWidget.itemSelectionChanged.connect((self.handleSelectionChanged))
                
        self.refreshTime.valueChanged.connect(self._refresh_time_handler)
        self.updateButton.released.connect(self.update_file)
        
        self.FileButton.clicked.connect(self.open_file)
        self.liveCheckBox.clicked.connect(self.live_update_onoff)
        self.pw_refresh_signal.connect(self.update_file)

    def closeEvent(self, event):
        widgetList = QApplication.topLevelWidgets()

        self.DATA._remove_plot_widgets( closeAll = True)
        self.DATA.set_info_thread_continue(False)
        event.accept()
    
    @pyqtSlot()
    def _close_plot_window(self):
        self.DATA._remove_plot_widgets()
                
    def live_update_onoff(self):
        if self.liveCheckBox.isChecked():
            self.timer.start(int(self.refreshTime_value))
            self.heardBeat.click()
        else:
            self.timer.stop()
    
    def _disable_live_update(self):
        try:
            if not self._force_live_plot and not self.h5file['entry'].attrs.get('updating',True):
                self.liveCheckBox.setChecked(False)
                self._force_live_plot = True # don't disable the live plot if it is manually enabled again
        except ValueError as e:
            qkit.logging.error("Qviewkit/DatasetsWindow.py: Error on checking for live plot: "+str(e))

    def _refresh_time_handler(self,refreshTime):
        self.refreshTime_value = refreshTime*1000 # ms -> s
        
    def populate_data_list(self):
        """
        populate_data_list is called regularly withing the refresh cycle to
        update the data tree.
        """
        self.parent = self.treeWidget.invisibleRootItem()
        column = 0
        parents = []
        s=""
        """itterate over the whole entry tree and collect the attributes """
        for i,pentry in enumerate(self.h5file["/entry"].keys()):
            tree_key = "/entry/"+pentry
            if tree_key not in self.DATA.ds_tree_items:
                parent = self.addParent(self.parent, column, str(pentry))
                self.DATA.ds_tree_items[tree_key] = parent
                parents.append(parent)
            else:
                parent = self.DATA.ds_tree_items[tree_key]
                
            s= "comment:\t"+str3(self.h5file[tree_key].attrs.get('comment',""))+"\n"
            self.DATA.dataset_info[tree_key] = s
            
            for j,centry in enumerate(self.h5file[tree_key].keys()):
                tree_key = "/entry/"+pentry+"/"+centry
                if tree_key not in self.DATA.ds_tree_items:
                    item = self.addChild(parent, column, str(centry),tree_key)
                    self.DATA.ds_tree_items[tree_key] = item
                    
                    if tree_key in self.DATA.ds_cmd_open:
                        # the order of the following two lines are important! otherwise two plots are opened
                        self.DATA.append_plot(self,item,tree_key)
                        item.setCheckState(0,QtCore.Qt.Checked)
                        self.update_plots()
                       
                s = ""
                try:
                    s="shape\t"+str(self.h5file[tree_key].shape)+"\n"
                    for k in list(self.h5file[tree_key].attrs.keys()):
                        s += k + "\t" + str3(self.h5file[tree_key].attrs[k]) + "\n"
                except TypeError:
                    s="shape\t"+str(self.h5file[tree_key].shape)+"\n"
                    for k in list(self.h5file[tree_key].attrs.keys()): 
                        s += k + "\t" + str(self.h5file[tree_key].attrs[k]) + "\n"
                except ValueError as e:
                    print("catch: populate data list:",e)
                
                self.DATA.dataset_info[tree_key] = s
               
    def addParent(self, parent, column, title,data = ''):
        item = QtGui.QTreeWidgetItem(parent, [title])
        #item.setData(column, QtCore.Qt.UserRole, data)
        item.setChildIndicatorPolicy(QtGui.QTreeWidgetItem.ShowIndicator)
        item.setExpanded (True)
        return item

    def addChild(self, parent, column, title, ds, data=''):
        item = QtGui.QTreeWidgetItem(parent, [title])
        item.setData(column, QtCore.Qt.UserRole, data)
        item.setCheckState (column, QtCore.Qt.Unchecked)
        return item

    def handleChanged(self, item, column):
        ds = str("/entry/"+item.parent().text(column)+"/"+item.text(column))
        if item.checkState(column) == QtCore.Qt.Checked:
            
            if not self.DATA.plot_is_open(ds):
                "this condition occures when the ds is opened from the cmd-l"
                self.DATA.append_plot(self,item,ds)

            # this is a not very sound hack of a concurrency problem!
            # better to use a locking technique            
            if not self.h5file:
                self.update_file()
            else:
                self.update_plots()
                
            # uncheck 
            window = self.DATA.open_plots[str(item)]
            window.destroyed.connect(self._close_plot_window)
                
        if item.checkState(column) == QtCore.Qt.Unchecked:
            
            if self.DATA.plot_is_open(ds):
                self.DATA.remove_plot(str(item),ds)
        
    def handleSelectionChanged(self):
        getSelected = self.treeWidget.selectedItems()
        if getSelected[0].parent():
            ds =str("/entry/"+getSelected[0].parent().text(0)+"/"+getSelected[0].text(0)) 
            self.Dataset_properties.clear()
            self.Dataset_properties.insertPlainText(self.DATA.dataset_info[ds])
        else:
            ds =str("/entry/"+getSelected[0].text(0)) 
            self.Dataset_properties.clear()
            self.Dataset_properties.insertPlainText(self.DATA.dataset_info[ds])
 
            
    def update_file(self):
        "update_file is regularly called when _something_ has to be updated. open-> do something->close"
        try:
            self.h5file= h5py.File(str(self.DATA.DataFilePath),mode='r')
            self.DATA.filename = self.h5file.filename.split(os.path.sep)[-1]
            self.populate_data_list()
            self.update_plots()
            self._disable_live_update()
            self.h5file.close()
            
            s = (self.DATA.DataFilePath.split(os.path.sep)[-5:])
            self.statusBar().showMessage((os.path.sep).join(s for s in s))
            
            title = "Qviewkit: %s"%(self.DATA.DataFilePath.split(os.path.sep)[-1:][0][:6])
            self.setWindowTitle(title)
        except IOError as e:
            print(e)

    def open_file(self):
        if in_pyqt5:
            _DataFilePath=str(QFileDialog.getOpenFileName(filter="*.h5")[0])
        if in_pyqt4:
            _DataFilePath=str(QFileDialog.getOpenFileName(filter="*.h5"))
            
        if _DataFilePath:
            self.DATA.DataFilePath = _DataFilePath
            self.h5file= h5py.File(self.DATA.DataFilePath,mode='r')
            self.DATA.filename = self.h5file.filename.split(os.path.sep)[-1]
            self.populate_data_list()
            self.h5file.close()
            
            s = (self.DATA.DataFilePath.split(os.path.sep)[-5:])
            self.statusBar().showMessage((os.path.sep).join(s for s in s))
            title = "Qviewkit: %s"%(self.DATA.DataFilePath.split(os.path.sep)[-1:][0][:6])
            self.setWindowTitle(title)
            
    def update_plots(self):
        self.refresh_signal.emit()


from qkit.core.instrument_base import Instrument
from tip_client import tip_client

class virtual_temperature_UFO(Instrument):
    '''
        this is real hardware connected to the experiment
        add one of those to automatically note its value in the measurement files
    '''

    def __init__(self, name):
        Instrument.__init__(self, name, tags=['virtual'])
        
        self.add_parameter('temperature', type=float,
            flags=Instrument.FLAG_GET, units='mK')
        
        self.t = tip_client('tip',address='pi-us74')   #tip raspberry instance
        
    def do_get_temperature(self):
        return round(float(self.t.r_get_T())*1000,2)   #T in mK



# Keithley 2636A class, to perform the communication between the Wrapper and the device
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
#
# based on the work by
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import time
import logging
import numpy

class Keithley_2636A(Instrument):
    '''
    This is the driver for the Keithley 2636A Source Meter

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Keithley_2636A', address='<GBIP address>, reset=<bool>')
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Keithley_2636A, and communicates with the wrapper.

        Input:
          name (string)    : name of the instrument
          address (string) : GPIB address
          reset (bool)     : resets to default values, default=False
        '''
        logging.info(__name__ + ' : Initializing instrument Keithley_2636A')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        
        self.add_parameter('voltageA',
            flags=Instrument.FLAG_GETSET, units='V', minval=-200, maxval=200, type=float)
        self.add_parameter('voltageB',
            flags=Instrument.FLAG_GETSET, units='V', minval=-200, maxval=200, type=float)
 
        self.add_parameter('currentA',
            flags=Instrument.FLAG_GETSET, units='A', minval=-10, maxval=10, type=float)
        self.add_parameter('currentB',
            flags=Instrument.FLAG_GETSET, units='A', minval=-10, maxval=10, type=float)
         
        self.add_parameter('statusA',
            flags=Instrument.FLAG_GETSET, type=bool)
        self.add_parameter('statusB',
            flags=Instrument.FLAG_GETSET, type=bool)

        self.add_function('reset')
        self.add_function('onA')
        self.add_function('offA')
        self.add_function('onB')
        self.add_function('offB')
 
        self.add_function('exec_tsl_script')
        self.add_function('exec_tsl_script_with_return')
        self.add_function('get_tsl_script_return')
        
        #self.add_function ('get_all')

        #self._visainstrument.write('beeper.beep(0.5,1000)')
        if (reset):
            self.reset()
        else:
            self.get_all()

    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : resetting instrument')
        self._visainstrument.write("reset()")
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : get all')
        #self.do_get_voltageA()
        #self.get_voltageB()
        #self.get_currentA()
        #self.get_currentB()
    
    def exec_tsl_script(self,script):
        '''
        Writes the script to the device
        Input:
            TSL script (basically lua)
        Output:
            None
        '''
        logging.debug(__name__ + ' : execute TSL script on device')
        self._visainstrument.write(script)
        
    def exec_tsl_script_with_return(self,script):
        '''
        Writes the script to the device
        Input:
            TSL script (basically lua)
        Output:
            None
        '''
        logging.debug(__name__ + ' : execute TSL script on device and return data')
        return self._visainstrument.ask(script).strip()

 
    def get_tsl_script_return(self):
        '''
        Writes the script to the device
        Input:
            None
        Output:
            data, can be anything
        '''
        logging.debug(__name__ + ' : return data from executed TSL script')
        return self._visainstrument.read().strip()

 
    def do_get_voltageA(self):
        '''
        Reads the phase of the signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get voltage A')
        return float(self._visainstrument.ask('print(smua.measure.v())').strip())

    def do_set_voltageA(self, volt):
        '''
        Set the Amplitude of the signal

        Input:
            volt (float) : voltage in volt

        Output:
            None
        '''
        logging.debug(__name__ + ' : set voltage of channel A to %f' % volt)
        self._visainstrument.write('smua.source.levelv = %s' % volt)

    def do_get_voltageB(self):
        '''
        Reads the Amplitude of the signal from the instrument

        Input:
            None

        Output:
            volt (float) : Voltage in Volts
        '''
        logging.debug(__name__ + ' : get voltage A')
        return float(self._visainstrument.ask('print(smub.measure.v())').strip())

    def do_set_voltageB(self, volt):
        '''
        Set the Amplitude of the signal

        Input:
            volt (float) : voltage in volt

        Output:
            None
        '''
        logging.debug(__name__ + ' : set voltage of channel B to %f' % volt)
        self._visainstrument.write('smub.source.levelv = %s' % volt)

    def do_get_currentA(self):
        '''
        Reads the phase of the signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + ' : get current A')
        return float(self._visainstrument.ask('print(smua.measure.i())').strip())

    def do_set_currentA(self, curr):
        '''
        Set the Amplitude of the signal

        Input:
            curr (float) : current in amps

        Output:
            None
        '''
        logging.debug(__name__ + ' : set current of channel A to %f' % curr)
        self._visainstrument.write('smua.source.leveli = %s' % curr)

    def do_get_currentB(self):
        '''
        Reads the phase of the signal from the instrument

        Input:
            None

        Output:
            current (float) : current in amps
        '''
        logging.debug(__name__ + ' : get current B')
        return float(self._visainstrument.ask('print(smub.measure.i())').strip())

    def do_set_currentB(self, curr):
        '''
        Set the Amplitude of the signal

        Input:
            curr (float) : current in amps

        Output:
            None
        '''
        logging.debug(__name__ + ' : set current of channel B to %f' % curr)
        self._visainstrument.write('smub.source.leveli = %s' % curr)

    def do_get_statusA(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (boolean)
        '''
        logging.debug(__name__ + ' : get status A')
        try:
            return int(float(self._visainstrument.ask('print(smua.source.output)').strip()))
        except:
            raise ValueError('Output status not specified : %s' % stat)
            return

    def do_set_statusA(self, status):
        '''
        Set the output status of the instrument

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : set status A to %s' %status)
        self._visainstrument.write('smua.source.output = %d' %status)

    def do_get_statusB(self):
        '''
        Reads the output status from the instrument

        Input:
            None

        Output:
            status (boolean) 
        '''
        try:
            return int(float(self._visainstrument.ask('print(smua.source.output)').strip()))
        except:
            raise ValueError('Output status not specified : %s' % stat)
            return

    def do_set_statusB(self, status):
        '''
        Set the output status of the instrument

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : set status A to %s' % status)
        self._visainstrument.write('smua.source.output = %d' %status)

    # shortcuts
    def offA(self):
        '''
        Set status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_statusA(False)

    def onA(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_statusA(True)

    def offB(self):
        '''
        Set status to 'off'

        Input:
            None

        Output:
            None
        '''
        self.set_statusB(False)

    def onB(self):
        '''
        Set status to 'on'

        Input:
            None

        Output:
            None
        '''
        self.set_statusB(True)



# Agilent DSO oscilloscope driver for qtlab
# Markus Jerger <markus.jerger@kit.edu>, 2011
# modified by Sebastian Probst <Sebastian.Probst@kit.edu>, 2012

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import numpy as np

class Agilent_DSO(Instrument):
    '''
    This driver is for Agilent digital storage oscilloscopes, in particular the
    DSO81204B 12GHz 40GS/s 4ch oscilloscope
    '''
    
    def __init__(self, name, address):
        '''
        
        Input:
            name (string)
            address (string): VISA address of the device
        '''
        
        Instrument.__init__(self, name, tags=['physical'])
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        
        # number of channels
        self._numchs = 4
        
        # initialize virtual instrument
        self.add_parameter('points', flags = Instrument.FLAG_GETSET, 
            units = '', type = int, minval = 16, maxval = 524288)
        self.add_parameter('segments', flags = Instrument.FLAG_GETSET, 
            units = '', type = int)
        self.add_parameter('averages', flags = Instrument.FLAG_GETSET, 
            units = '', type = int, minval = 1, maxval = 4096)
        self.add_parameter('average_enabled', flags = Instrument.FLAG_GETSET, 
            units = '', type = bool)
                    
        self.add_parameter('samplerate', flags = Instrument.FLAG_GETSET, 
            units = 'Hz', type = float)
        self.add_parameter('bandwidth', flags = Instrument.FLAG_GETSET, 
            units = 'Hz', type = float)
        self.add_parameter('acqmode', flags = Instrument.FLAG_GET, 
            type = str)
        self.add_function('acqmode_realtime')
        self.add_function('acqmode_peakdetect')
        self.add_function('acqmode_highres')
        self.add_function('acqmode_segmented')
            
        self.add_parameter('refclock', flags = Instrument.FLAG_GETSET,
            type = bool)
        self.add_parameter('xoffset', flags = Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            units = 's', type = float)
        self.add_parameter('xrange', flags = Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            units = 's', type = float)
        
        self.add_parameter('yoffset', flags = Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET, 
            units = 'V', type = float, channels=(1, self._numchs), channel_prefix='ch%d_')
        self.add_parameter('yrange', flags = Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET, 
            units = 'V', type = float, channels=(1, self._numchs), channel_prefix='ch%d_')
            
        self.add_function('reset')
        self.add_function('preset')
        self.add_function('clear_status')
        self.add_function('get_error')
        self.add_function('message')
        self.add_function('acq_digitize')
        self.add_function('acq_run')
        self.add_function('acq_single')
        self.add_function('acq_stop')
        self.add_function('get_triggered')
        self.add_function('get_done')
        self.add_function('get_completedpercent')
        self.add_function('get_averages_acquired')
        self.add_function('autoscale')
        self.add_function('get_data_old')
        self.add_function('get_data')
        self.add_function('restart_average')
        self.add_function('get_completed')
        self.add_function('get_meas_phase')
        self.add_function('get_meas_vamp')
        self.add_function('get_meas_vrms')
        self.add_function('get_meas_frequency')
        
        # to do: functions to setup maths on the oscilloscope
        # to do: histogram functions
        # to do: marker stuff
        # to do: measurements
        # to do: trigger setup functions

        # initialize physical instrument
        self._visainstrument.write(':SYST:HEAD OFF')
        self._visainstrument.write(':SYST:LONG OFF')
        self._visainstrument.write(':WAV:BYT LSBF') # faster
            
    # memory management: points, segments and averages
    def do_set_average_enabled(self,value):
        if value: self._visainstrument.write(':ACQ:AVER ON')
        else: self._visainstrument.write(':ACQ:AVER Off')
        
    def do_get_average_enabled(self):
        if 1 == self._visainstrument.ask(':ACQ:AVER?'): value=True
        else: value = False
        return value
    
    def do_get_points(self):
        return int(self._visainstrument.ask(':ACQ:POIN?'))
    def do_set_points(self, value):
        self._visainstrument.write(':ACQ:POIN %d'%value)
    def do_get_segments(self):
        return int(self._visainstrument.ask(':ACQ:SEGM:COUN?'))
    def do_set_segments(self, value):
        self._visainstrument.write(':ACQ:SEGM:COUN %d'%value)
    def do_get_averages(self):
        avgon = 1 == int(self._visainstrument.ask(':ACQ:AVER?'))
        if(avgon):
            return int(self._visainstrument.ask(':ACQ:AVER:COUN?'))
        else:
            return 1
    def do_set_averages(self, value):
        if(value > 1):
            self._visainstrument.write(':ACQ:AVER 1')
            self._visainstrument.write(':ACQ:AVER:COUN %d'%value)
        else:
            self._visainstrument.write(':ACQ:AVER 0')


    # acquisition options bandwidth
    def do_get_samplerate(self):
        srate = self._visainstrument.ask(':ACQ:SRATE?')
        if(srate == 'AUTO'):
            return 0
        else:
            return float(srate)
    def do_set_samplerate(self, value):
        if(value == 0): value = 'AUTO'
        if(value == inf): value = 'MAX'
        self._visainstrument.write(':ACQ:BAND %s'%string(value))    
    def do_get_bandwidth(self):
        band = self._visainstrument.ask(':ACQ:BAND?')
        if(band == 'AUTO'):
            return inf
        else:
            return float(band)
    def do_set_bandwidth(self, value):
        if value in [0,inf] : value = 'AUTO'   ##((value == 0) || (value == inf))
        self._visainstrument.write(':ACQ:BAND %s'%string(value))
    def do_get_acqmode(self):
        '''
            acquisition mode may be RTIM, PDET, HRES, SEGM
        '''
        return self._visainstrument.ask(':ACQ:MODE?')
    def acqmode_realtime(self):
        self._visainstrument.write(':ACQ:MODE RTIM')
    def acqmode_peakdetect(self):
        self._visainstrument.write(':ACQ:MODE PDET')
    def acqmode_highres(self):
        self._visainstrument.write(':ACQ:MODE HRES')
    def acqmode_segmented(self):
        self._visainstrument.write(':ACQ:MODE SEGM')
    
    # timebase functions
    def do_get_xoffset(self,):
        '''
            return the delay between trigger event and the delay reference point
        '''
        return float(self._visainstrument.ask(':TIM:POS?'))
    def do_set_xoffset(self, value):
        return self._visainstrument.write(':TIM:POS %e'%value)
    def do_get_xrange(self,):
        '''
            return the delay between trigger event and the delay reference point
        '''
        return float(self._visainstrument.ask(':TIM:RANG?'))
    def do_set_xrange(self, value):
        return self._visainstrument.write(':TIM:RANG %e'%value)    
    def do_get_refclock(self):
        '''
            use external 10MHz reference?
        '''
        return '1' == self._visainstrument.ask(':TIM:REFC?')
    def do_set_refclock(self, value):
        return self._visainstrument.write(':TIM:REFC %d'%int(value))
        
    # per channel functions
    def do_get_yoffset(self, channel):
        return float(self._visainstrument.ask(':CHAN%d:OFFS?'%channel))
    def do_set_yoffset(self, value, channel):
        return self._visainstrument.write(':CHAN%d:OFFS %e'%(channel, value))
    def do_get_yrange(self, channel):
        return float(self._visainstrument.ask(':CHAN%d:RANG?'%channel))
    def do_set_yrange(self, value, channel):
        return self._visainstrument.write(':CHAN%d:RANG %e'%(channel, value))
        
    # functions
    def reset(self):
        self._visainstrument.write('*RST')
    def preset(self):
        self._visainstrument.write(':PRES')
    def autoscale(self):
        self._visainstrument.write(':AUT')
    def clear_status(self):
        '''
            clear status bits and empty error queue
        '''
        self._visainstrument.write('*CLS')
    def get_error(self):
        '''
            return error code and message of the next error in the queue
            if the queue is empty or the last entry was reached, returns a 0 error code
            
            Returns:
                (int) error code, (string) error message
        '''
        reply = self._visainstrument.ask(':SYST:ERR? STR')
        errno, null, errmsg = str.partition(reply)
        return int(errno), errmsg[1:-1]
    def message(self, value):
        '''
            display a message on the advisory line
        '''
        self._visainstrument.write(':SYST:DSP %s'%str(value))
    def acq_digitize(self, channels = None, functions = None):
        '''
            runs the instument in digitizer mode
            measurement and math operations will not be executed
            (see chapter 6 of the programming manual for an example)
            Input:
                channels/functions - numbers of the channels or functions to acquire
        '''
        channels = ['CHAN%d'%x for x in channels]
        functions = ['FUNC%d'%x for x in functions]
        parameter = ', '.join(channels + functions)
        self._visainstrument.write(':DIG %s'%parameter)
    def acq_run(self):
        '''
            runs the instrument in standard mode
        '''
        self._visainstrument.write('*TRG')
    def acq_single(self):
        '''
            runs the instrument in single acquisition mode
        '''
        self._visainstrument.write(':SING')
    def acq_stop(self):
        '''
            stops the instrument
        '''
        self._visainstrument.write(':STOP')
    def get_triggered(self):
        '''
            returns True if the instrument was triggered since the last call
        '''
        return '1' == self._visainstrument.ask(':TER?')
    def get_done(self):
        '''
            returns True if the acquisition has finished
        '''
        return 1 == self._visainstrument.ask(':ADER?')
    def get_completedpercent(self):
        return float(self._visainstrument.ask(':WAV:COMP?'))
        
    def get_completed(self):
        return float(self._visainstrument.ask(':ACQ:COMP?'))

    def get_averages_acquired(self):
        return int(self._visainstrument.ask(':WAV:COUNT?'))
        
    def restart_average(self):
        self.do_set_average_enabled(False)
        self.do_set_average_enabled(True)
        
    def get_data_old(self, channel):
        '''
            retrieve waveform data from the oscilloscope
            
            Input:
                source - CHAN<N>, FUNC<N>, HIST, WMEM<N>
            Returns:
                data as a numpy array of floats
                preamble sent by the device
        '''
        # retrieve data preamble
        self._visainstrument.write(':WAV:VIEW ALL; :WAV:FORM WORD')
        values = self._visainstrument.ask(':WAV:PRE?')#.split(',')
        print len(values)
        print values
        values = values.split(',')
        print values
        def conv_to_float_if_possible(val):
            for x in val:
                if x.replace('.','',1).isdigit()==True : x=float(x)
            return val
        values = [conv_to_float_if_possible(value) for value in values]
        parts = [
            'format', 'type', 'points', 'count', 
            'Xincrement', 'Xorigin', 'Xreference',
            'Yincrement', 'Yorigin', 'Yreference', 'coupling', 
            'Xdisprange', 'Xdisporigin', 'Ydisprange', 'Ydisporigin',
            'date', 'time', 'framemodelno', 'acqmode', 'completion',
            'Xunits', 'Yunits', 'lowerBWlimit', 'upperBWlimit'
        ]
        print len(parts)
        print len(values)
        print values
        preamble = {}
        for i in range(len(parts)):
            preamble.update({parts[i]:values[i]})  #this is very strange values[0][i]
        ##preamble = dict(parts=values)
        
        
        # retrieve data from device
        self._visainstrument.write(':WAV:SOUR %s'%channel)
        raw_data = self._visainstrument.ask_for_values(':WAV:DATA?', '') # start,size; format ascii, single, double, big_endian
        ##print np.shape(raw_data)
        ##print raw_data
        ##data = np.fromstring(raw_data, np.int16) ###??? not float?
        data = 0
        
        return data, preamble
        
    def get_data(self, channel, format='ascii'): #format='word' doesn't work!!!
        '''
            retrieve waveform data from the oscilloscope
            
            Input:
                source - CHAN<N>, FUNC<N>, HIST, WMEM<N>
            Returns:
                data as a numpy array of floats
                preamble sent by the device
        '''
        self._visainstrument.write(':WAV:SOUR CHAN%i' %channel)
        self._visainstrument.write(':WAV:BYTEORDER LSBF')
        if format=='ascii': self._visainstrument.write(':WAV:FORM ASCii')
        if format=='word': self._visainstrument.write(':WAV:FORM WORD')
        xUnits = self._visainstrument.ask(':WAV:XUN?')
        xOrigin = float(self._visainstrument.ask(':WAV:XOR?'))
        xIncrement = float(self._visainstrument.ask(':WAV:XINC?'))
        #print xIncrement
        xPoints = int(self._visainstrument.ask(':WAV:POIN?'))
        #print xPoints
        
        xData = [xOrigin+xIncrement*i for i in range(xPoints)]
        
        yOrigin = float(self._visainstrument.ask(':WAV:YOR?'))
        yIncrement = float(self._visainstrument.ask(':WAV:YINC?'))
        if format=='word':
            yData = self._visainstrument.ask_for_values(':WAV:DATA?',format=1)
            yData = [yOrigin+yIncrement*float(i) for i in yData]
        if format=='ascii': yData = [float(i) for i in self._visainstrument.ask(':WAV:DATA?').split(',')]
        return xData, yData
        
    #measurement functions
    def get_meas_phase(self, channel1, channel2):
        return float(self._visainstrument.ask(':MEAS:PHAS? CHAN%d,CHAN%d'%(channel1,channel2)))
        
    def get_meas_vamp(self, channel1):
        return float(self._visainstrument.ask(':MEAS:VAMP? CHAN%d'%channel1))

    def get_meas_vrms(self,channel1,display=True,ac=True):
        '''
            Get RMS Voltage value of channel1
            display: Calculate RMS over whole display (true) or single cycle (false)
            AC: Remove DC component (true) or include it (false)
        '''        
        return float(self._visainstrument.ask(':MEAS:VRMS? %s,%s,CHAN%d'%('DISP' if display else 'CYCL','AC' if ac else 'DC',channel1)))
        
    def get_meas_area(self,channel1,display=True):
        '''
            Get RMS Voltage value of channel1
            display: Calculate RMS over whole display (true) or single cycle (false)
        '''        
        return float(self._visainstrument.ask(':MEAS:AREA? %s,CHAN%d'%('DISP' if display else 'CYCL',channel1)))
    
    def get_meas_frequency(self, channel1):
        return float(self._visainstrument.ask(':MEAS:FREQ? CHAN%d'%channel1))    
    
    def get_meas(self):
        return self._visainstrument.ask(':MEAS:RES?')
    
    def set_meas_statistics(self, parameter):
        self._visainstrument.write(':MEAS:STAT %s'%parameter)
    
    def ask(self, question):
        return self._visainstrument.ask('%s'%question)
    def write(self, question):
        return self._visainstrument.write('%s'%question)



# -*- coding: utf-8 -*-
"""
@author: hannes.rotzinger@kit.edu / 2015,2016,2017 
         marco.pfirrmann@kit.edu / 2016, 2017
Qlicense: GPL
"""
# support both PyQt4 and 5
in_pyqt5 = False
import sys
try:
    from PyQt5 import QtCore
    from PyQt5.QtCore import Qt,QObject,pyqtSlot
    from PyQt5.QtWidgets import QWidget,QPlainTextEdit,QMenu,QAction, QWidgetAction, QLabel, QActionGroup
    from PyQt5.QtGui import QBrush, QPainter, QPixmap, QFont
    #from PyQt5 import Qt
    in_pyqt5 = True
except ImportError as e:
    pass
if not in_pyqt5:
    try:
        from PyQt4 import QtCore
        from PyQt4.QtCore import Qt,QObject,pyqtSlot
        from PyQt4.QtGui import QWidget,QPlainTextEdit,QMenu,QAction, QWidgetAction, QLabel, QBrush, QPainter, QPixmap, QActionGroup, QFont
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)

import pyqtgraph as pg
import numpy as np

import qkit
from qkit.gui.qviewkit.plot_view import Ui_Form
from qkit.storage.hdf_constants import ds_types, view_types
from qkit.gui.qviewkit.PlotWindow_lib import _display_1D_view, _display_1D_data, _display_2D_data, _display_table, _display_text
from qkit.gui.qviewkit.PlotWindow_lib import _get_ds, _get_ds_url, _get_name, _get_unit
from qkit.core.lib.misc import str3

class PlotWindow(QWidget,Ui_Form):
    """PlotWindow class organizes the correct display of data in a h5 file.

    This class coordinates the display of the data between the UI with it's 
    selector slots in Ui_Form and the plot functions in PlotWindow_lib. Here we
    handle the trigger events; update, open, and close the plot windows.
    Depending on the selector slot settings the associated plot function in the
    lib is called.
    We inherit from PyQt's QWidget class as well as the Ui_Form class.
    
    Heart and soul here is update_plots() which is connected to a timer trigger
    and checks the settings of all user slots and calls the correct lib
    function to display the data. The window starts with some default
    configuration for the dataset types.
    """
    def myquit(self):
        exit()

    def __init__(self,parent,data,dataset_url):
        """Inits PlotWindow with a parent (QMainWindow.treeWidget), data
        (qviewkit.main.DATA object), and a dataset_url to be opened.
        """
        self.DATA = data
        self.dataset_url = dataset_url
        self.obj_parent = parent

        super(PlotWindow , self).__init__()
        Ui_Form.__init__(self)
        # move forward ...
        self.setWindowState(Qt.WindowActive)
        self.activateWindow()
        self.raise_()

        "This variable controlles if a window is new, see update_plots()."
        self._windowJustCreated = True
        "connect update_plots to the DatasetWindow"
        self.obj_parent.refresh_signal.connect(self.update_plots)

    def closeEvent(self, event):
        "overwrite the closeEvent handler"
        self.DATA._toBe_deleted(self.dataset_url)
        self.DATA._remove_plot_widgets()
        event.accept()

    @pyqtSlot()
    def update_plots(self):
        """Checks the signal slots and changes the plotting attributes to be 
        evaluated by PlotWindow_lib.
        
        This function coordinates the changing input from the signal slots and
        updates the attributes that are parsed to the plotting lib functions.
        update_plots() is either periodically called e.g. by the timer or once 
        on startup.
        
        Args:
            self: Object of the PlotWindow class.
        Returns:
            No return variable. The function operates on the given object.
        """
        #print "PWL update_plots:", self.obj_parent.h5file
        
        try:
            self.ds = self.obj_parent.h5file[self.dataset_url]
        except ValueError as e:
            print(str(self.dataset_url)+": "+str(e))
            return
        self.ds_type = self.ds.attrs.get('ds_type', -1)
        
        # The axis names are parsed to plot_view's Ui_Form class to label the UI selectors 
        x_ds_name = _get_name(_get_ds(self.ds, self.ds.attrs.get('x_ds_url', '')))
        y_ds_name = _get_name(_get_ds(self.ds, self.ds.attrs.get('y_ds_url', '')))
        z_ds_name = _get_name(_get_ds(self.ds, self.ds.attrs.get('z_ds_url', '')))
        try:
            if self.ds.attrs.get('xy_0', ''):
                x_ds_name_view = _get_name(_get_ds(self.ds,_get_ds(self.ds, str3(self.ds.attrs.get('xy_0', '')).split(':')[1]).attrs.get('x_ds_url','')))
                y_ds_name_view = _get_name(_get_ds(self.ds,_get_ds(self.ds, str3(self.ds.attrs.get('xy_0', '')).split(':')[1]).attrs.get('y_ds_url','')))
            else:
                raise AttributeError
        except AttributeError:
            x_ds_name_view = '_none_'
            y_ds_name_view = '_none_'

        selector_label = [x_ds_name, y_ds_name, z_ds_name, x_ds_name_view, y_ds_name_view]    

        ## At first window creation the attributes are set to state zero.
        if self._windowJustCreated:
            # A few state variables:
            self._onPlotTypeChanged = True
            self._windowJustCreated = False

            self.graphicsView = None
            self.TraceValueChanged  = False
            self.TraceZValueChanged = False
            self.TraceXValueChanged = False
            self.TraceYValueChanged = False
            self.VTraceZValueChanged = False
            self.VTraceXValueChanged = False
            self.VTraceYValueChanged = False

            # the following calls rely on ds_type and setup the layout of the plot window.
            self.setupUi(self,self.ds_type, selector_label)

            window_title = str(self.dataset_url.split('/')[-1]) +" "+ str(self.DATA.filename)
            self.setWindowTitle(window_title)

            self._setDefaultView()
            self._setup_signal_slots()
        
        ## We check here for the view type given by either default setting or
        ## user input. Creates the canvas and calls the associated plot 
        ## function from the lib. For code constistence the variable names are
        ## the same in all view types.
        try:
            if self.view_type == view_types['1D-V']:
                if not self.graphicsView or self._onPlotTypeChanged:
                    self._onPlotTypeChanged = False
                    self.graphicsView = pg.PlotWidget(name=self.dataset_url)
                    self.graphicsView.setObjectName(self.dataset_url)
                    self.addQvkMenu()
                    self.graphicsView.sceneObj.contextMenu[:0] = [self.qvkMenu]
                    self.gridLayout.addWidget(self.graphicsView,0,0)
                _display_1D_view(self,self.graphicsView)

            elif self.view_type == view_types['1D']:
                if not self.graphicsView or self._onPlotTypeChanged:
                    self._onPlotTypeChanged = False
                    self.graphicsView = pg.PlotWidget(name=self.dataset_url)
                    self.graphicsView.setObjectName(self.dataset_url)
                    self.addQvkMenu()
                    self.graphicsView.sceneObj.contextMenu[:0] = [self.qvkMenu]
                    self.gridLayout.addWidget(self.graphicsView,0,0)
                _display_1D_data(self,self.graphicsView)

            elif self.view_type == view_types['2D']:
                if not self.graphicsView or self._onPlotTypeChanged:
                    self._onPlotTypeChanged = False
                    self.graphicsView = ImageViewMplColorMaps(self.obj_parent, view = pg.PlotItem())
                    self.graphicsView.setObjectName(self.dataset_url)
                    self.addQvkMenu()
                    self.graphicsView.scene.contextMenu[:0] = [self.qvkMenu]
                    self.graphicsView.view.setAspectLocked(False)
                    self.gridLayout.addWidget(self.graphicsView,0,0)
                    self.linestyle_selector.setDisabled(True)
                _display_2D_data(self,self.graphicsView)

            elif self.view_type == view_types['table']:
                if not self.graphicsView or self._onPlotTypeChanged:
                    self._onPlotTypeChanged = False
                    self.graphicsView = pg.TableWidget(sortable=False)
                    self.graphicsView.setWindowTitle(self.dataset_url+'_table')
                    self.graphicsView.setObjectName(self.dataset_url)
                    self.graphicsView.setFormat("%.6g")
                    self.gridLayout.addWidget(self.graphicsView,0,0)
                _display_table(self,self.graphicsView)

            elif self.view_type == view_types['txt']:
                if not self.graphicsView or self._onPlotTypeChangeBox:
                    self._onPlotTypeChangeBox = False
                    self.graphicsView = QPlainTextEdit()
                    self.graphicsView.setObjectName(self.dataset_url)
                    self.gridLayout.addWidget(self.graphicsView,0,0)
                self.graphicsView.clear()
                _display_text(self,self.graphicsView)
            else:
                print("This should not be here: View Type:"+str(self.view_type))
        except ValueError as e:
            print("PlotWindow: Value Error; Dataset not yet available", self.dataset_url)
            print(e)
        except OSError as e:
            print("PlotWindow: OS Error; Probably a read/write conflict occured", self.dataset_url)
            print(e)


    def _setup_signal_slots(self):
        """Depending on the dataset type the possible signal slots are created
        
        Args:
            self: Object of the PlotWindow class.
        Returns:
            No return variable. The function operates on the given object.
        """
        if self.ds_type == ds_types['vector'] or self.ds_type == ds_types['coordinate']:
            self.PlotTypeSelector.currentIndexChanged.connect(self._onPlotTypeChangeVector)

        elif self.ds_type == ds_types['matrix'] or self.ds_type == -1:
            self.PlotTypeSelector.currentIndexChanged.connect(self._onPlotTypeChangeMatrix)
            self.TraceXSelector.valueChanged.connect(self._setBTraceXNum)
            self.TraceYSelector.valueChanged.connect(self._setBTraceYNum)
            self.TraceXValue.returnPressed.connect(self._setBTraceXValue)            
            self.TraceYValue.returnPressed.connect(self._setBTraceYValue) 

        elif self.ds_type == ds_types['box']:
            self.PlotTypeSelector.currentIndexChanged.connect(self._onPlotTypeChangeBox)
            self.TraceXSelector.valueChanged.connect(self._setBTraceXNum)
            self.TraceYSelector.valueChanged.connect(self._setBTraceYNum)
            self.TraceZSelector.valueChanged.connect(self._setBTraceZNum)
            self.TraceXValue.returnPressed.connect(self._setBTraceXValue)            
            self.TraceYValue.returnPressed.connect(self._setBTraceYValue)            
            self.TraceZValue.returnPressed.connect(self._setBTraceZValue)

        elif self.ds_type == ds_types['view']:
            self.VTraceXSelector.valueChanged.connect(self._setVTraceXNum)
            self.VTraceYSelector.valueChanged.connect(self._setVTraceYNum)
            self.VTraceXValue.returnPressed.connect(self._setVTraceXValue)           
            self.VTraceYValue.returnPressed.connect(self._setVTraceYValue)

    def keyPressEvent(self, ev):
        #print("Received Key Press Event!! You Pressed: "+ event.text())
        if ev.key() == Qt.Key_S:
            #print '# ',ev.key()
            print(self.data_coord)
            sys.stdout.flush()

    def _setDefaultView(self):
        """Set the display attributes to a ds_type sensitive default case.
        
        The default attributs for the ds_types are:
        coordinate -> 1d, data vs. x_coord
        vector -> 1d, data vs. x_coord
        matrix -> 2d, color coded data with y_coord and x_coord
        box -> 2d, color coded data with y_coord and x_coord at the mid-point
            of the z-coord.
        The default settings for the user signal slots are set up in their 
        respective functions.
        Args:
            self: Object of the PlotWindow class.
        Returns:
            No return variable. The function operates on the given object.
        """
        
        ## Some look-up dicts for manipulation and plot styles.
        self.plot_styles = {'line':0,'linepoint':1,'point':2}
        self.manipulations = {'dB':1, 'wrap':2, 'linear':4, 'remove_zeros':8,
                              'sub_offset_avg_x':16, 'sub_offset_avg_y':32, 'norm_data_avg_x':64, 'norm_data_avg_y':128, 'histogram':256} #BITMASK for manipulation

        self.plot_style = 0
        self.manipulation = 8
        self.TraceNum = -1
        self.view_type = self.ds.attrs.get("view_type",None)
        self.user_setting_changed = False

        if self.ds_type == ds_types["coordinate"]:
            self.view_type = view_types['1D']
            self._defaultCoord()

        elif self.ds_type == ds_types["vector"]:
            self.view_type = view_types['1D']
            self._defaultVector()

        elif self.ds_type == ds_types["matrix"]:
            self.view_type = view_types['2D']
            self._defaultMatrix()

        elif self.ds_type == ds_types["box"]:
            self.view_type = view_types['2D']
            self._defaultBox()

        elif self.ds_type == ds_types["txt"]:
            self.view_type = view_types['txt']

        elif self.ds_type == ds_types["view"]:
            self.view_type = view_types['1D-V']
            self._defaultView()

        else:
            self._defaultOld()

    def _defaultCoord(self):
        self.PlotTypeSelector.setCurrentIndex(0)

    def _defaultVector(self):
        self.PlotTypeSelector.setCurrentIndex(0)

    def _defaultMatrix(self):
        shape = self.ds.shape
        self.TraceXSelector.setEnabled(False)
        self.TraceXSelector.setRange(-1*shape[0],shape[0]-1)
        self.TraceXNum = -1        

        self.TraceYSelector.setEnabled(False)
        self.TraceYSelector.setRange(-1*shape[1],shape[1]-1)
        self.TraceYNum = -1
        
        self.PlotTypeSelector.setCurrentIndex(0)

    def _defaultBox(self):
        shape = self.ds.shape
        
        self.TraceZSelector.setEnabled(True)
        self.TraceZSelector.setRange(-1*shape[2],shape[2]-1)
        self.TraceZSelector.setValue(int(shape[2]/2))
        self.TraceZNum = int(shape[2]/2)
        
        self.TraceXSelector.setEnabled(False)
        self.TraceXSelector.setRange(-1*shape[0],shape[0]-1)
        self.TraceXNum = -1        

        self.TraceYSelector.setEnabled(False)
        self.TraceYSelector.setRange(-1*shape[1],shape[1]-1)
        self.TraceYNum = -1
        
        self.PlotTypeSelector.setCurrentIndex(2)

    def _defaultView(self):
        self.VTraceXSelector.setEnabled(True)
        self.VTraceYSelector.setEnabled(True)
        self.VTraceXNum = -1
        self.VTraceYNum = -1 

    def _defaultOld(self):
        if not self.view_type:
            if len(self.ds.shape) == 1:
                self.PlotTypeSelector.setCurrentIndex(1)
                self.view_type = view_types['1D']
                self.TraceSelector.setEnabled(False)
                self.PlotTypeSelector.setEnabled(False)
                self.plot_style = self.plot_styles['line']
            elif len(self.ds.shape) == 2:
                self.TraceSelector.setEnabled(False)
                shape = self.ds.shape[0]
                self.TraceSelector.setRange(-1*shape,shape-1)
                self.PlotTypeSelector.setCurrentIndex(0)
                self.view_type = view_types['2D']
            elif len(self.ds.shape) == 3:
                self.TraceSelector.setEnabled(False)
                shape = self.ds.shape[0]
                self.view_type = view_types['2D']
                self.PlotTypeSelector.setEnabled(False)
            else:
                self.TraceSelector.setEnabled(True)
                self.view_type = view_types['1D']
        else:
            self.TraceSelector.setEnabled(True)
            self.PlotTypeSelector.setEnabled(False)


    def _setTraceNum(self,num):
        self.TraceNum = num
        if not self.TraceValueChanged:
            self.obj_parent.pw_refresh_signal.emit()
    #######
    def _setVTraceXValue(self):
        xval = str(self.VTraceXValue.displayText())
        try:
            self._VtraceX_value = float(xval.split()[0])
        except ValueError:
            return
        self.VTraceXValueChanged = True
        self.obj_parent.pw_refresh_signal.emit()


    def _setVTraceXNum(self,num):
        self.VTraceXNum = num
        if not self.VTraceXValueChanged:
            self.obj_parent.pw_refresh_signal.emit()

    def _setVTraceYValue(self):
        xval = str(self.VTraceYValue.displayText())
        try:
            self._VtraceY_value = float(xval.split()[0])
        except ValueError:
            return
        self.VTraceYValueChanged = True
        self.obj_parent.pw_refresh_signal.emit()


    def _setVTraceYNum(self,num):
        self.VTraceYNum = num
        if not self.TraceYValueChanged:
            self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceXValue(self):
        xval = str(self.TraceXValue.displayText())
        try:
            self._traceX_value = float(xval.split()[0])
        except ValueError:
            return
        self.TraceXValueChanged = True
        self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceXNum(self,num):
        self.TraceXNum = num
        if not self.TraceXValueChanged:
            self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceYValue(self):
        xval = str(self.TraceYValue.displayText())
        try:
            self._traceY_value = float(xval.split()[0])
        except ValueError:
            return
        self.TraceYValueChanged = True
        self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceYNum(self,num):
        self.TraceYNum = num
        if not self.TraceYValueChanged:
            self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceZValue(self):
        xval = str(self.TraceZValue.displayText())
        try:
            self._traceZ_value = float(xval.split()[0])
        except ValueError:
            return
        self.TraceZValueChanged = True
        self.obj_parent.pw_refresh_signal.emit()

    def _setBTraceZNum(self,num):
        self.TraceZNum = num
        if not self.TraceZValueChanged:
            self.obj_parent.pw_refresh_signal.emit()

    def _onPlotTypeChangeVector(self, index):
        self._onPlotTypeChanged = True
        if index == 0:
            self.view_type = view_types['1D']
        if index == 1:
            self.view_type = view_types['table']

        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    def _onPlotTypeChangeMatrix(self, index):
        self._onPlotTypeChanged = True
        if index == 0:
            self.view_type = view_types['2D']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(False)
        if index == 1:
            self.view_type = view_types['1D']
            self.TraceXSelector.setEnabled(True)
            self.TraceYSelector.setEnabled(False)
        if index == 2:
            self.view_type = view_types['1D']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(True)
        if index == 3:
            self.view_type = view_types['table']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(False)

        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    def _onPlotTypeChangeBox(self, index):
        self._onPlotTypeChanged = True
        if index == 0:
            self.view_type = view_types['2D']
            self.TraceXSelector.setEnabled(True)
            self.TraceYSelector.setEnabled(False)
            self.TraceZSelector.setEnabled(False)
        if index == 1:
            self.view_type = view_types['2D']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(True)
            self.TraceZSelector.setEnabled(False)
        if index == 2:
            self.view_type = view_types['2D']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(False)
            self.TraceZSelector.setEnabled(True)
        if index == 3:
            self.view_type = view_types['1D']
            self.TraceXSelector.setEnabled(False)
            self.TraceYSelector.setEnabled(True)
            self.TraceZSelector.setEnabled(True)
        if index == 4:
            self.view_type = view_types['1D']
            self.TraceXSelector.setEnabled(True)
            self.TraceYSelector.setEnabled(False)
            self.TraceZSelector.setEnabled(True)
        if index == 5:
            self.view_type = view_types['1D']
            self.TraceXSelector.setEnabled(True)
            self.TraceYSelector.setEnabled(True)
            self.TraceZSelector.setEnabled(False)

        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    def _addXPlotChange(self):
        pass
    def _addYPlotChange(self):
        pass

    def _init_XY_add(self):
        for i,key in enumerate(self.DATA.ds_tree_items.keys()):
            keys = key.split("/")[-2:]
            key = "/".join(key for key in keys)
            self.addXPlotSelector.addItem("")
            self.addYPlotSelector.addItem("")
            self.addXPlotSelector.setItemText(i, str(key))
            self.addYPlotSelector.setItemText(i, str(key))
            #print i,key



    def _getXValueFromTraceNum(self,ds,num):
        x_ds = _get_ds(ds, ds.attrs.get('x_ds_url'))
        if x_ds is None:
            return "__none__"
        if ds.attrs.get('ds_type') == ds_types['box']:
            x_data = np.array(x_ds)[:ds.attrs.get('fill')[0]]
        else:
            x_data = np.array(x_ds)[:ds.shape[0]]
        xunit = _get_unit(x_ds)
        try:
            xval = x_data[num]
        except (KeyError,IndexError):
            xval = "X"
        return str(xval)+" "+str(xunit)

    def _getYValueFromTraceNum(self,ds,num):
        y_ds = _get_ds(ds, ds.attrs.get('y_ds_url'))
        if ds.attrs.get('ds_type') == ds_types['box']:
            y_data = np.array(y_ds)[:ds.attrs.get('fill')[1]]
        else:
            y_data = np.array(y_ds)[:ds.shape[1]]
        yunit = _get_unit(y_ds)
        try:
            yval = y_data[num]
        except (KeyError,IndexError):
            yval = "X"
        return str(yval)+" "+str(yunit)

    def _getZValueFromTraceNum(self,ds,num):
        z_ds = _get_ds(ds, ds.attrs.get('z_ds_url'))
        z_data = np.array(z_ds)[:ds.shape[2]]
        zunit = _get_unit(z_ds)
        try:
            zval = z_data[num]
        except (KeyError,IndexError):
            zval = "X"
        return str(zval)+" "+str(zunit)
        

    def addQvkMenu(self,menu=None):
        """Add custom entry in the right-click menu.
        
        The data manipulation option are displayed in the menu. Each entry is
        connected with a trigger event and the information about the clicked
        entry is parsed to the plot lib via the plot_style or the manipulation
        attribute.
        Args:
            self: Object of the PlotWindow class.
            menu: Menu of the respective pyqtgraph plot class
        Returns:
            No return variable. The function operates on the given object.
        """
    
        self.qvkMenu = QMenu("Qviewkit")

        self.linestyle_selector = QActionGroup(self.qvkMenu)
        self.linestyle_selector.setExclusive(True)

        self.linestyle_selector.point = QAction('Point', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(self.linestyle_selector.point)
        self.linestyle_selector.point.triggered.connect(self.setPointMode)
        self.linestyle_selector.addAction(self.linestyle_selector.point)

        self.linestyle_selector.line = QAction('Line', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(self.linestyle_selector.line)
        self.linestyle_selector.line.triggered.connect(self.setLineMode)
        self.linestyle_selector.addAction(self.linestyle_selector.line)

        self.linestyle_selector.pointLine = QAction('Point+Line', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(self.linestyle_selector.pointLine)
        self.linestyle_selector.pointLine.triggered.connect(self.setPointLineMode)
        self.linestyle_selector.addAction(self.linestyle_selector.pointLine)
        
        self.qvkMenu.addSeparator()
        
        phase_wrap = QAction('(un)wrap phase data', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(phase_wrap)
        phase_wrap.triggered.connect(self.setPhaseWrap)
        
        linear_correction = QAction('linearly correct data', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(linear_correction)
        linear_correction.triggered.connect(self.setLinearCorrection)

        offset_correction_x = QAction('data-<data:x>', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(offset_correction_x)
        offset_correction_x.triggered.connect(self.setOffsetCorrectionX)
        
        offset_correction_y = QAction('data-<data:y>', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(offset_correction_y)
        offset_correction_y.triggered.connect(self.setOffsetCorrectionY)

        norm_correction_x = QAction('data/<data:x>', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(norm_correction_x)
        norm_correction_x.triggered.connect(self.setNormDataCorrectionX)
        
        norm_correction_y = QAction('data/<data:y>', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(norm_correction_y)
        norm_correction_y.triggered.connect(self.setNormDataCorrectionY)

        dB_scale = QAction('dB / linear', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(dB_scale)
        dB_scale.triggered.connect(self.setdBScale)

        histogram = QAction('Histogram', self.qvkMenu, checkable=True)
        self.qvkMenu.addAction(histogram)
        histogram.triggered.connect(self.setHistogram)

        self.manipulation_menu = {'dB': dB_scale,
                                  'wrap': phase_wrap,
                                  'linear': linear_correction,
                                  'sub_offset_avg_x': offset_correction_x,
                                  'sub_offset_avg_y': offset_correction_y,
                                  'norm_data_avg_x': norm_correction_x,
                                  'norm_data_avg_y': norm_correction_y,
                                  'histogram': histogram}
        
        self.updateManipulationMenu()
        
        self.qvkMenu.addSeparator()
        
        notefont = QFont("Helvetica",7,12)
        note = QAction("Manipulations are executed in the order shown here.",self.qvkMenu,enabled=False,font=notefont)
        self.qvkMenu.addAction(note)
        
        self.qvkMenu.setTearOffEnabled(True)
        self.qvkMenu.setWindowTitle("Qviewkit "+str(self.dataset_url.split('/')[-1]) +" "+ str(self.DATA.filename))
        
        if menu is not None:
            menu.addMenu(self.qvkMenu)
    
    def updateManipulationMenu(self):
        for m in self.manipulation_menu:
            self.manipulation_menu[m].setChecked(bool(self.manipulation & self.manipulations[m]))

    @pyqtSlot()
    def setPointMode(self):
        self.plot_style = self.plot_styles['point']
        self.user_setting_changed = True
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setLineMode(self):
        self.plot_style = self.plot_styles['line']
        self.user_setting_changed = True
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setPointLineMode(self):
        self.plot_style = self.plot_styles['linepoint']
        self.user_setting_changed = True
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setdBScale(self):
        self.manipulation = self.manipulation ^ self.manipulations['dB']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()
      
    @pyqtSlot()
    def setPhaseWrap(self):
        self.manipulation = self.manipulation ^ self.manipulations['wrap']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setLinearCorrection(self):
        self.manipulation = self.manipulation ^ self.manipulations['linear']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()
            
    @pyqtSlot()
    def setOffsetCorrectionX(self):
        self.manipulation = self.manipulation ^ self.manipulations['sub_offset_avg_x']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setOffsetCorrectionY(self):
        self.manipulation = self.manipulation ^ self.manipulations['sub_offset_avg_y']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setNormDataCorrectionX(self):
        self.manipulation = self.manipulation ^ self.manipulations['norm_data_avg_x']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setNormDataCorrectionY(self):
        self.manipulation = self.manipulation ^ self.manipulations['norm_data_avg_y']
        #print "{:08b}".format(self.manipulation)
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()

    @pyqtSlot()
    def setHistogram(self):
        self.manipulation = self.manipulation ^ self.manipulations['histogram']
        if not self._windowJustCreated:
            self.obj_parent.pw_refresh_signal.emit()


from collections import OrderedDict

class ImageViewMplColorMaps(pg.ImageView):
    """Class to manually inclued the matplotlib 2.0 colormaps.
    
    The class implements the matplotlib 2.0 colormaps 'viridis', 'inferno', 
    'plasma', and 'magma' to the available colors scales in our 2d plots.
    They are not (yet?) available in the standard pyqtgraph package. However, 
    there is a pull request #446, issued Feb 2017 by github user "WFrsh" to add
    them to the standard available gradient dict.
    The code is a combination by his added rgb color maps and a monkey patch by
    user "honkomonk" posted in the comment section of pull request #561.
    
    The color gradients for the maps are hard coded here and added to the
    available dict of colormaps. The 'viridis' map is set as default.
    """
    def __init__(self, parent=None, name='ImageView', view=None, imageItem=None, *args):
        super(ImageViewMplColorMaps, self).__init__(parent=parent, name=name, view=view, imageItem=imageItem, *args)

        self.gradientEditorItem = self.ui.histogram.item.gradient

        self.activeCm = "viridis"
        self.mplColorMaps = OrderedDict([
                    ('inferno', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (87, 15, 109, 255)), (0.5, (187, 55, 84, 255)), (0.75, (249, 142, 8, 255)), (1.0, (252, 254, 164, 255))], 'mode': 'rgb'}),
                    ('plasma', {'ticks': [(0.0, (12, 7, 134, 255)), (0.25, (126, 3, 167, 255)), (0.5, (203, 71, 119, 255)), (0.75, (248, 149, 64, 255)), (1.0, (239, 248, 33, 255))], 'mode': 'rgb'}),
                    ('magma', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (80, 18, 123, 255)), (0.5, (182, 54, 121, 255)), (0.75, (251, 136, 97, 255)), (1.0, (251, 252, 191, 255))], 'mode': 'rgb'}),
                    ('viridis', {'ticks': [(0.0, (68, 1, 84, 255)), (0.25, (58, 82, 139, 255)), (0.5, (32, 144, 140, 255)), (0.75, (94, 201, 97, 255)), (1.0, (253, 231, 36, 255))], 'mode': 'rgb'}),
                    ])

        self.registerCmap()

    def registerCmap(self):
        """ Add matplotlib cmaps to the GradientEditors context menu"""
        self.gradientEditorItem.menu.addSeparator()
        savedLength = self.gradientEditorItem.length
        self.gradientEditorItem.length = 100
        
        
        for name in self.mplColorMaps:
            px = QPixmap(100, 15)
            p = QPainter(px)
            self.gradientEditorItem.restoreState(self.mplColorMaps[name])
            grad = self.gradientEditorItem.getGradient()
            brush = QBrush(grad)
            p.fillRect(QtCore.QRect(0, 0, 100, 15), brush)
            p.end()
            label = QLabel()
            label.setPixmap(px)
            label.setContentsMargins(1, 1, 1, 1)
            act =QWidgetAction(self.gradientEditorItem)
            act.setDefaultWidget(label)
            act.triggered.connect(self.cmapClicked)
            act.name = name
            self.gradientEditorItem.menu.addAction(act)
        self.gradientEditorItem.length = savedLength


    def cmapClicked(self):
        """onclick handler for our custom entries in the GradientEditorItem's context menu"""
        act = self.sender()
        self.gradientEditorItem.restoreState(self.mplColorMaps[act.name])
        self.activeCm = act.name